from enum import Enum
from django.utils.functional import cached_property
from rest_framework import serializers
from arkindex.project.serializer_fields import EnumField
from arkindex.images.models import ImageServer
from arkindex.documents.models import ElementType
import fnmatch
import yaml


class ImportType(Enum):
    Volumes = 'volumes'
    Acts = 'acts'
    Metadata = 'metadata'


class VolumesImportFormat(Enum):
    IIIF = 'iiif'
    TXT = 'txt'


class ActsImportFormat(Enum):
    CSV = 'csv'


class MetadataImportFormat(Enum):
    TEI = 'tei'


class ImportConfig(dict):

    def path_match(self, path):
        return any(fnmatch.fnmatch(path, pattern) for pattern in self['paths'])


class ImportConfigSerializer(serializers.Serializer):
    format = serializers.CharField()
    paths = serializers.ListField(
        child=serializers.CharField(),
        allow_empty=False,
        min_length=1,
        read_only=False,
    )

    default_error_messages = {
        'type_folder': 'Element type with slug {slug!r} in {name!r} should be a folder',
        'type_hidden': 'Element type with slug {slug!r} in {name!r} should be hidden',
        'type_not_found': 'Element type with slug {slug!r} in {name!r} not found',
        'type_not_folder': 'Element type with slug {slug!r} in {name!r} should not be a folder',
        'type_not_hidden': 'Element type with slug {slug!r} in {name!r} should not be hidden',
    }

    class Meta:
        output_class = ImportConfig

    def validate(self, data):
        return self.Meta.output_class(data)


class VolumesConfig(ImportConfig):

    @cached_property
    def image_servers(self):
        if not self.get('image_servers'):
            return list(ImageServer.objects.all())

        servers = []
        for name, url in self['image_servers'].items():
            s, _ = ImageServer.objects.get_or_create(url=url, defaults={'display_name': name})
            servers.append(s)
        return servers


class VolumesConfigSerializer(ImportConfigSerializer):
    format = EnumField(VolumesImportFormat, default=VolumesImportFormat.IIIF)
    lazy_checks = serializers.BooleanField(default=False)
    autoconvert_https = serializers.BooleanField(default=False)
    image_servers = serializers.DictField(child=serializers.URLField(), default=dict)
    manifest_type = serializers.SlugField(default='volume')
    canvas_type = serializers.SlugField(default='page')

    def validate_manifest_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='manifest_type')

        if not element_type.folder:
            self.fail('type_folder', slug=slug, name='manifest_type')
        if element_type.hidden:
            self.fail('type_not_hidden', slug=slug, name='manifest_type')

        return element_type

    def validate_canvas_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='canvas_type')

        if element_type.folder:
            self.fail('type_not_folder', slug=slug, name='canvas_type')
        if element_type.hidden:
            self.fail('type_not_hidden', slug=slug, name='canvas_type')

        return element_type

    class Meta:
        output_class = VolumesConfig


class ActsConfigSerializer(ImportConfigSerializer):
    format = EnumField(ActsImportFormat, default=ActsImportFormat.CSV)
    folder_type = serializers.SlugField(default='volume')
    act_type = serializers.SlugField(default='act')
    surface_type = serializers.SlugField(default='surface')

    def validate_folder_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='folder_type')

        if not element_type.folder:
            self.fail('type_folder', slug=slug, name='folder_type')
        if element_type.hidden:
            self.fail('type_not_hidden', slug=slug, name='folder_type')

        return element_type

    def validate_act_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='act_type')

        if element_type.folder:
            self.fail('type_not_folder', slug=slug, name='act_type')
        if element_type.hidden:
            self.fail('type_not_hidden', slug=slug, name='act_type')

        return element_type

    def validate_surface_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='surface_type')

        if element_type.folder:
            self.fail('type_not_folder', slug=slug, name='surface_type')
        if not element_type.hidden:
            self.fail('type_hidden', slug=slug, name='surface_type')

        return element_type


class MetadataConfigSerializer(ImportConfigSerializer):
    format = EnumField(MetadataImportFormat, default=MetadataImportFormat.TEI)
    folder_type = serializers.SlugField(default='volume')
    element_type = serializers.SlugField(default='act')

    def validate_element_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='element_type')
        return element_type

    def validate_folder_type(self, slug):
        assert 'corpus' in self.context, 'A corpus is required to validate element types'
        try:
            element_type = self.context['corpus'].types.get(slug=slug)
        except ElementType.DoesNotExist:
            self.fail('type_not_found', slug=slug, name='folder_type')

        if not element_type.folder:
            self.fail('type_folder', slug=slug, name='folder_type')
        if element_type.hidden:
            self.fail('type_not_hidden', slug=slug, name='folder_type')

        return element_type


class GitConfig(dict):
    FILE_NAME = '.arkindex.yml'

    def branch_match(self, branch_name):
        if branch_name.startswith('refs/heads/'):
            branch_name = branch_name[11:]
        return branch_name in self['branches']


class GitConfigSerializer(serializers.Serializer):
    """
    Serializer for .arkindex.yml configuration files
    """

    version = serializers.IntegerField(min_value=1, max_value=1)
    branches = serializers.ListField(
        child=serializers.CharField(),
        allow_empty=False,
        min_length=1,
        read_only=False,
    )
    volumes = VolumesConfigSerializer(required=False)
    acts = ActsConfigSerializer(required=False)
    metadata = MetadataConfigSerializer(required=False)

    default_error_messages = {
        'no_imports': 'No import types were set in the configuration file',
    }

    @classmethod
    def from_path(cls, path):
        with open(path, 'rb') as f:
            return cls(data=yaml.safe_load(f))

    def validate(self, data):
        if not any(import_type.value in data for import_type in ImportType):
            self.fail('no_imports')
        return GitConfig(data)
