from abc import ABC, abstractmethod
from arkindex.dataimport.config import GitConfig, VolumesImportFormat
from arkindex.dataimport.iiif import IIIFParser
from arkindex.documents.acts import ActsImporter
from arkindex.documents.tei import TeiParser
import os.path
import logging


logger = logging.getLogger(__name__)


class FileType(ABC):
    """
    A file type that can be handled by Git import workflows
    """

    @staticmethod
    def get(flow, diff):
        """
        Return a FileType class corresponding to a given file path, or raise ValueError if it is unknown.
        """
        try:
            filetype = next((ft for ft in file_types if ft.match(diff.new_path, flow.config)))
        except StopIteration:
            raise ValueError('Unknown file type for "{}"'.format(diff.new_path))
        return filetype(flow, diff)

    @staticmethod
    def sort(instances):
        """
        Sort FileType instances by priority
        """
        return sorted(instances, key=lambda f: file_types.index(f.__class__))

    @staticmethod
    @abstractmethod
    def match(path, config):
        """
        Returns True if the specified path matches this file type.
        """

    def __init__(self, flow, diff):
        from arkindex.dataimport.git import GitFlow
        assert isinstance(flow, GitFlow)
        assert self.match(diff.new_path, flow.config), \
            '{} cannot handle {}'.format(self.__class__.__name__, diff.new_path)
        assert self.match(diff.old_path, flow.config), 'File type changes are not supported'
        self.flow = flow
        self.diff = diff
        # List of updated elements
        self.changed_elements = []

    def __str__(self):
        return '{} - {} from {} to {}'.format(
            self.__class__.__name__,
            self.diff.type.name,
            self.diff.old_path,
            self.diff.new_path,
        )

    @property
    def full_path(self):
        return os.path.join(self.flow.repo_dir, self.diff.new_path)

    @abstractmethod
    def handle(self):
        """
        Handle a Git diff on a single file. Should run synchronously.
        """


class IIIFFileType(FileType):
    """
    IIIF manifests describing volumes and pages, with the volume name as their file name
    """

    @staticmethod
    def match(path, config):
        assert isinstance(config, GitConfig)
        return config.get('volumes') and \
            config['volumes']['format'] == VolumesImportFormat.IIIF and \
            config['volumes'].path_match(path) and \
            path.endswith('.json')

    def handle(self):
        # TODO: Handle a pure deletion diff
        self.changed_elements.extend(IIIFFileType.run(self.flow.dataimport, self.full_path, self.flow.config))

    @classmethod
    def run(cls, dataimport, path, config):
        parser = IIIFParser.from_dataimport(dataimport, path, config)
        return parser.run()


class ManifestListFileType(FileType):
    """
    Lists of IIIF manifest URLs in text files
    """

    @staticmethod
    def match(path, config):
        assert isinstance(config, GitConfig)
        return config.get('volumes') and \
            config['volumes']['format'] == VolumesImportFormat.TXT and \
            config['volumes'].path_match(path) and \
            path.endswith('.txt')

    def handle(self):
        # TODO: Handle a pure deletion diff
        # TODO: Maybe look at the actual Git diff to see which manifests in the list were modified?
        with open(self.full_path) as f:
            paths = [line.strip() for line in f.read().splitlines()]

        for path in paths:
            logger.info('Parsing manifest {}'.format(path))
            self.changed_elements.extend(IIIFFileType.run(self.flow.dataimport, path, self.flow.config))


class ActsListFileType(FileType):
    """
    CSV files with a volume name as their filename and two columns: [act number, first folio]
    """

    @staticmethod
    def match(path, config):
        assert isinstance(config, GitConfig)
        return config.get('acts') and config['acts'].path_match(path) and path.endswith('.csv')

    def handle(self):
        ActsImporter(
            self.full_path,
            corpus=self.flow.dataimport.corpus,
            folder_type=self.flow.config['acts']['folder_type'],
            act_type=self.flow.config['acts']['act_type'],
            surface_type=self.flow.config['acts']['surface_type'],
        ).run()


class MetadataFileType(FileType):
    """
    TEI-XML files
    """

    @staticmethod
    def match(path, config):
        assert isinstance(config, GitConfig)
        return config.get('metadata') and config['metadata'].path_match(path) and path.endswith('.xml')

    def handle(self):
        parser = TeiParser(
            path=self.full_path,
            corpus=self.flow.dataimport.corpus,
            revision=self.flow.dataimport.revision,
            folder_type=self.flow.config['metadata']['folder_type'],
            element_type=self.flow.config['metadata']['element_type'],
        )
        parser.run()


class GitConfigFileType(FileType):
    """
    An Arkindex repository configuration file
    """

    @staticmethod
    def match(path, config):
        return os.path.basename(path) == GitConfig.FILE_NAME

    def handle(self):
        # Do nothing (config files are already handled in every other import)
        return


# Registered file types for the FileType.get static method, ordered by priority
file_types = [
    GitConfigFileType,
    ManifestListFileType,
    IIIFFileType,
    ActsListFileType,
    MetadataFileType,
]
