from rest_framework.exceptions import ValidationError, ErrorDetail
from arkindex.project.tests import FixtureTestCase
from arkindex.images.models import ImageServer
from arkindex.dataimport.config import \
    GitConfigSerializer, VolumesImportFormat, ActsImportFormat, MetadataImportFormat


class TestGitConfig(FixtureTestCase):
    """
    Tests for the .arkindex.yml configuration file parser
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.sample_server = ImageServer.objects.create(display_name='sample', url='http://example.com/iiif/')

    def setUp(self):
        # A valid configuration; tests will edit/remove values to check for validations
        self.base_data = {
            "version": 1,
            "branches": ["master", "dev"],
            "volumes": {
                "format": "txt",
                "paths": ["volumes/*", "manifests/*"],
                "image_servers": {
                    "sample": "http://example.com/iiif/",
                },
                "lazy_checks": True,
                "autoconvert_https": True,
            },
            "acts": {
                "format": "csv",
                "paths": ["acts/*"],
            },
            "metadata": {
                "format": "tei",
                "paths": ["metadata/*"],
            }
        }

    def test_base(self):
        """
        Test parsing a valid configuration file
        """
        cfg = GitConfigSerializer(data=self.base_data, context={'corpus': self.corpus})
        cfg.is_valid(raise_exception=True)
        data = cfg.validated_data
        self.assertEqual(data['version'], 1)
        self.assertListEqual(data['branches'], ['master', 'dev'])

        volumes = data['volumes']
        self.assertEqual(volumes['format'], VolumesImportFormat.TXT)
        self.assertListEqual(volumes['paths'], ['volumes/*', 'manifests/*'])
        self.assertTrue(volumes['lazy_checks'])
        self.assertTrue(volumes['autoconvert_https'])
        self.assertEqual(volumes['manifest_type'], self.corpus.types.get(slug='volume'))
        self.assertEqual(volumes['canvas_type'], self.corpus.types.get(slug='page'))
        self.assertEqual(volumes.image_servers, [self.sample_server])

        acts = data['acts']
        self.assertEqual(acts['format'], ActsImportFormat.CSV)
        self.assertListEqual(acts['paths'], ['acts/*'])
        self.assertEqual(acts['act_type'], self.corpus.types.get(slug='act'))
        self.assertEqual(acts['surface_type'], self.corpus.types.get(slug='surface'))

        metadata = data['metadata']
        self.assertEqual(metadata['format'], MetadataImportFormat.TEI)
        self.assertListEqual(metadata['paths'], ['metadata/*'])
        self.assertEqual(metadata['folder_type'], self.corpus.types.get(slug='volume'))
        self.assertEqual(metadata['element_type'], self.corpus.types.get(slug='act'))

    def test_version(self):
        """
        Test the GitConfigSerializer handles versioning
        """
        self.base_data['version'] = 42  # Bad version number
        with self.assertRaisesRegex(ValidationError, 'version'):
            GitConfigSerializer(
                data=self.base_data,
                context={'corpus': self.corpus},
            ).is_valid(raise_exception=True)

    def test_branches(self):
        """
        Test the GitConfigSerializer handles branches
        """
        self.base_data['branches'] = ['this', 'is', 'a', {'bad': 'format'}]
        with self.assertRaisesRegex(ValidationError, 'branches'):
            GitConfigSerializer(
                data=self.base_data,
                context={'corpus': self.corpus},
            ).is_valid(raise_exception=True)

    def test_no_imports(self):
        """
        Test the config parser requires at least one import
        """
        del self.base_data['volumes']
        del self.base_data['acts']
        del self.base_data['metadata']
        with self.assertRaisesRegex(ValidationError, 'import types'):
            GitConfigSerializer(
                data=self.base_data,
                context={'corpus': self.corpus},
            ).is_valid(raise_exception=True)

    def test_import_paths(self):
        """
        Test the config parser checks for valid paths
        """
        del self.base_data['volumes']['paths']
        with self.assertRaisesRegex(ValidationError, 'paths'):
            GitConfigSerializer(
                data=self.base_data,
                context={'corpus': self.corpus},
            ).is_valid(raise_exception=True)

        self.base_data['volumes']['paths'] = ['bad', {'paths': 'format'}]
        with self.assertRaisesRegex(ValidationError, 'paths'):
            GitConfigSerializer(
                data=self.base_data,
                context={'corpus': self.corpus},
            ).is_valid(raise_exception=True)

    def test_import_formats(self):
        """
        Test omitting import formats sets to default values
        """
        del self.base_data['volumes']['format']
        del self.base_data['acts']['format']
        del self.base_data['metadata']['format']
        cfg = GitConfigSerializer(data=self.base_data, context={'corpus': self.corpus})
        cfg.is_valid(raise_exception=True)
        self.assertEqual(cfg.validated_data['volumes']['format'], VolumesImportFormat.IIIF)
        self.assertEqual(cfg.validated_data['acts']['format'], ActsImportFormat.CSV)
        self.assertEqual(cfg.validated_data['metadata']['format'], MetadataImportFormat.TEI)

    def test_types_exist(self):
        """
        Test setting non-existent type slugs causes errors
        """
        self.base_data['volumes']['manifest_type'] = 'bread'
        self.base_data['volumes']['canvas_type'] = 'lettuce'
        self.base_data['acts']['act_type'] = 'tomato'
        self.base_data['acts']['surface_type'] = 'bacon'
        self.base_data['metadata']['folder_type'] = 'steak'
        self.base_data['metadata']['element_type'] = 'mayonnaise'

        cfg = GitConfigSerializer(data=self.base_data, context={'corpus': self.corpus})
        self.assertFalse(cfg.is_valid())
        self.assertDictEqual(cfg.errors, {
            'volumes': {
                'manifest_type': [ErrorDetail(
                    string="Element type with slug 'bread' in 'manifest_type' not found",
                    code='type_not_found',
                )],
                'canvas_type': [ErrorDetail(
                    string="Element type with slug 'lettuce' in 'canvas_type' not found",
                    code='type_not_found',
                )],
            },
            'acts': {
                'act_type': [ErrorDetail(
                    string="Element type with slug 'tomato' in 'act_type' not found",
                    code='type_not_found',
                )],
                'surface_type': [ErrorDetail(
                    string="Element type with slug 'bacon' in 'surface_type' not found",
                    code='type_not_found',
                )],
            },
            'metadata': {
                'folder_type': [ErrorDetail(
                    string="Element type with slug 'steak' in 'folder_type' not found",
                    code='type_not_found',
                )],
                'element_type': [ErrorDetail(
                    string="Element type with slug 'mayonnaise' in 'element_type' not found",
                    code='type_not_found',
                )],
            }
        })

    def test_folders(self):
        """
        Test folder-only or non-folder types ensure the element types are or aren't folders
        """
        self.base_data['volumes']['manifest_type'] = 'page'
        self.base_data['volumes']['canvas_type'] = 'volume'
        self.base_data['acts']['act_type'] = 'volume'
        self.base_data['acts']['surface_type'] = 'volume'
        self.base_data['metadata']['folder_type'] = 'act'
        self.base_data['metadata']['element_type'] = 'volume'

        cfg = GitConfigSerializer(data=self.base_data, context={'corpus': self.corpus})
        self.assertFalse(cfg.is_valid())
        self.assertDictEqual(cfg.errors, {
            'volumes': {
                'manifest_type': [ErrorDetail(
                    string="Element type with slug 'page' in 'manifest_type' should be a folder",
                    code='type_folder',
                )],
                'canvas_type': [ErrorDetail(
                    string="Element type with slug 'volume' in 'canvas_type' should not be a folder",
                    code='type_not_folder',
                )],
            },
            'acts': {
                'act_type': [ErrorDetail(
                    string="Element type with slug 'volume' in 'act_type' should not be a folder",
                    code='type_not_folder',
                )],
                'surface_type': [ErrorDetail(
                    string="Element type with slug 'volume' in 'surface_type' should not be a folder",
                    code='type_not_folder',
                )],
            },
            'metadata': {
                'folder_type': [ErrorDetail(
                    string="Element type with slug 'act' in 'folder_type' should be a folder",
                    code='type_folder',
                )],
            }
        })

    def test_hidden(self):
        """
        Test hidden-only or non-hidden types ensure the element types are or aren't hidden
        """
        sneaky_volume = self.corpus.types.create(slug='sneaky_volume', hidden=True, folder=True)
        self.base_data['volumes']['manifest_type'] = sneaky_volume.slug
        self.base_data['volumes']['canvas_type'] = 'surface'
        self.base_data['acts']['act_type'] = 'surface'
        self.base_data['acts']['surface_type'] = 'page'
        self.base_data['metadata']['folder_type'] = sneaky_volume.slug
        self.base_data['metadata']['element_type'] = 'surface'

        cfg = GitConfigSerializer(data=self.base_data, context={'corpus': self.corpus})
        self.assertFalse(cfg.is_valid())
        self.assertDictEqual(cfg.errors, {
            'volumes': {
                'manifest_type': [ErrorDetail(
                    string="Element type with slug 'sneaky_volume' in 'manifest_type' should not be hidden",
                    code='type_not_hidden',
                )],
                'canvas_type': [ErrorDetail(
                    string="Element type with slug 'surface' in 'canvas_type' should not be hidden",
                    code='type_not_hidden',
                )],
            },
            'acts': {
                'act_type': [ErrorDetail(
                    string="Element type with slug 'surface' in 'act_type' should not be hidden",
                    code='type_not_hidden',
                )],
                'surface_type': [ErrorDetail(
                    string="Element type with slug 'page' in 'surface_type' should be hidden",
                    code='type_hidden',
                )],
            },
            'metadata': {
                'folder_type': [ErrorDetail(
                    string="Element type with slug 'sneaky_volume' in 'folder_type' should not be hidden",
                    code='type_not_hidden',
                )],
            }
        })
