from unittest.mock import patch, call, MagicMock
from arkindex_common.enums import MetaType, DataImportMode
from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.filetypes import \
    FileType, IIIFFileType, ActsListFileType, MetadataFileType, GitConfigFileType
from arkindex.dataimport.git import GitFlow, SimpleDiff, DiffType
from arkindex.dataimport.config import GitConfig, VolumesConfig, ImportConfig, VolumesImportFormat


class TestFileTypes(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = cls.corpus.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.di = cls.corpus.imports.create(
            mode=DataImportMode.Repository,
            creator=cls.user,
            revision=cls.rev,
            payload={
                'repo_id': str(cls.repo.id),
                'sha': cls.rev.hash,
            },
        )
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.surface_type = cls.corpus.types.get(slug='surface')
        with patch('arkindex.dataimport.git.os.path.isdir', lambda _: True):
            cls.flow = GitFlow(cls.di, 'the/')

    @patch('arkindex.dataimport.filetypes.file_types', [])
    def test_filetype_get_fail(self):
        with self.assertRaisesRegex(ValueError, 'Unknown file type'):
            FileType.get(self.flow, SimpleDiff(DiffType.Addition, 'a', 'a'))

    def test_manifest_match(self):
        self.flow.config = GitConfig({
            'volumes': VolumesConfig(format=VolumesImportFormat.IIIF, paths=['*']),
        })
        self.flow.config['volumes'].path_match = MagicMock(return_value=True)
        self.assertTrue(IIIFFileType.match('test.json', self.flow.config))

        self.assertEqual(self.flow.config['volumes'].path_match.call_count, 1)
        self.assertEqual(self.flow.config['volumes'].path_match.call_args, call('test.json'))

        self.assertFalse(IIIFFileType.match('test.xml', self.flow.config))

        self.flow.config['volumes'].path_match.return_value = False
        self.assertFalse(IIIFFileType.match('test.json', self.flow.config))

    def test_acts_match(self):
        self.flow.config = GitConfig({
            'acts': ImportConfig(paths=['*']),
        })
        self.flow.config['acts'].path_match = MagicMock(return_value=True)
        self.assertTrue(ActsListFileType.match('path/to/some.csv', self.flow.config))
        self.assertFalse(ActsListFileType.match('wrong/type.xml', self.flow.config))

        self.assertEqual(self.flow.config['acts'].path_match.call_count, 2)
        self.assertEqual(self.flow.config['acts'].path_match.call_args_list, [
            call('path/to/some.csv'),
            call('wrong/type.xml'),
        ])

        self.flow.config['acts'].path_match.return_value = False
        self.assertFalse(ActsListFileType.match('path/to/some.csv', self.flow.config))

    @patch('arkindex.dataimport.filetypes.ActsImporter')
    def test_acts_handle(self, importer_mock):
        self.flow.config = GitConfig({
            'acts': ImportConfig({
                'paths': ['*'],
                'folder_type': self.volume_type,
                'act_type': self.act_type,
                'surface_type': self.surface_type,
            }),
        })
        acts_diff = SimpleDiff(DiffType.Modification, 'a.csv', 'b.csv')
        ft = ActsListFileType(self.flow, acts_diff)
        ft.handle()
        self.assertEqual(importer_mock.call_count, 1)
        self.assertEqual(importer_mock.call_args, call(
            'the/repo/b.csv',
            corpus=self.corpus,
            folder_type=self.volume_type,
            act_type=self.act_type,
            surface_type=self.surface_type,
        ))
        self.assertEqual(importer_mock().run.call_count, 1)
        self.assertEqual(importer_mock().run.call_args, ())

    def test_config_match(self):
        self.assertTrue(GitConfigFileType.match('path/to/.arkindex.yml', self.flow.config))
        self.assertFalse(GitConfigFileType.match('path/to/somethingelse.yml', self.flow.config))

    def test_metadata_match(self):
        self.flow.config = GitConfig({
            'metadata': ImportConfig(paths=['*']),
        })
        self.flow.config['metadata'].path_match = MagicMock()

        self.assertTrue(MetadataFileType.match('path/to/some.xml', self.flow.config))
        self.assertFalse(MetadataFileType.match('wrong/type.csv', self.flow.config))

        self.assertEqual(self.flow.config['metadata'].path_match.call_count, 2)
        self.assertEqual(self.flow.config['metadata'].path_match.call_args_list, [
            call('path/to/some.xml'),
            call('wrong/type.csv'),
        ])

        self.flow.config['metadata'].path_match.return_value = False
        self.assertFalse(MetadataFileType.match('path/to/some.xml', self.flow.config))

    @patch('arkindex_common.tei.Date')
    @patch('arkindex.dataimport.filetypes.TeiParser')
    def test_metadata_handle(self, tei_parser_mock, tei_elt_mock):
        db_elt = self.corpus.elements.get(name='Act 1')
        metadata = db_elt.metadatas.create(
            name='date',
            type=MetaType.Date,
            value='1333',
        )

        tei_parser_mock().match_database.return_value = [(db_elt, tei_elt_mock())]
        tei_parser_mock().run.return_value = [metadata]

        self.flow.config = GitConfig({
            'metadata': ImportConfig({
                'paths': ['*'],
                'folder_type': self.volume_type,
                'element_type': self.act_type,
            }),
        })

        metadatas_diff = SimpleDiff(DiffType.Modification, 'a.xml', 'b.xml')

        ft = MetadataFileType(self.flow, metadatas_diff)
        ft.handle()

        self.assertEqual(tei_parser_mock.call_args, call(
            path='the/repo/b.xml',
            corpus=self.corpus,
            revision=self.rev,
            folder_type=self.volume_type,
            element_type=self.act_type,
        ))
        self.assertEqual(tei_parser_mock().run.call_count, 1)
