from unittest.mock import patch, MagicMock
from rest_framework.exceptions import APIException, NotAuthenticated, AuthenticationFailed, ValidationError
from gitlab.exceptions import GitlabGetError, GitlabCreateError
from ponos.models import State
from arkindex_common.enums import DataImportMode
from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.providers import GitLabProvider
from arkindex.dataimport.models import DataImport


class TestGitLabProvider(FixtureTestCase):
    """
    Test the GitLabProvider class
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.gl_patch = patch('arkindex.dataimport.providers.Gitlab')

    def setUp(self):
        super().setUp()
        self.gl_mock = self.gl_patch.start()

    def tearDown(self):
        super().tearDown()
        self.gl_patch.stop()

    def test_list_repos(self):
        """
        Test GitLabProvider can list repositories from GitLab
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).list_repos()

        self.assertEqual(self.gl_mock.call_count, 1)
        args, kwargs = self.gl_mock.call_args
        self.assertTupleEqual(args, ('http://aaa', ))
        self.assertDictEqual(kwargs, {'oauth_token': self.creds.token})

        self.assertEqual(self.gl_mock().projects.list.call_count, 1)
        args, kwargs = self.gl_mock().projects.list.call_args
        self.assertTupleEqual(args, ())
        self.assertDictEqual(kwargs, {'membership': True, 'search': None})

    def test_list_repos_query(self):
        """
        Test GitLabProvider can search repositories from GitLab
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).list_repos(query='meh')

        self.assertEqual(self.gl_mock.call_count, 1)
        args, kwargs = self.gl_mock.call_args
        self.assertTupleEqual(args, ('http://aaa', ))
        self.assertDictEqual(kwargs, {'oauth_token': self.creds.token})

        self.assertEqual(self.gl_mock().projects.list.call_count, 1)
        args, kwargs = self.gl_mock().projects.list.call_args
        self.assertTupleEqual(args, ())
        self.assertDictEqual(kwargs, {'membership': True, 'search': 'meh'})

    def test_list_repos_requires_credentials(self):
        """
        Test GitLabProvider checks for credentials when requesting repositories list
        """
        with self.assertRaises(NotAuthenticated):
            GitLabProvider(url='http://aaa').list_repos()

    def test_create_repo(self):
        """
        Test GitLabProvider can create a Repository instance from a GitLab repo
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        new_repo = GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
            id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        args, kwargs = self.gl_mock().projects.get.call_args
        self.assertTupleEqual(args, (1337, ))
        self.assertDictEqual(kwargs, {})

        self.assertEqual(new_repo.url, 'http://new_repo_url')
        self.assertEqual(new_repo.corpus, self.corpus)
        self.assertEqual(new_repo.provider_name, 'GitLabProvider')

        self.assertEqual(self.gl_mock().projects.get().hooks.create.call_count, 1)
        args, kwargs = self.gl_mock().projects.get().hooks.create.call_args
        self.assertEqual(len(args), 1)
        self.assertDictEqual(kwargs, {})
        self.assertDictEqual(args[0], {
            'url': 'http://hook',
            'push_events': True,
            'token': new_repo.hook_token,
        })

    def test_create_repo_requires_credentials(self):
        """
        Test GitLabProvider checks for credentials when requesting a repository creation
        """
        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        with self.assertRaises(NotAuthenticated):
            GitLabProvider(url='http://aaa').create_repo(
                id='repo_id', request=request_mock, corpus=self.corpus)

    def test_create_repo_already_exists(self):
        """
        Test GitLabProvider checks for duplicate repositories
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
            id='1337', request=request_mock, corpus=self.corpus)

        with self.assertRaises(ValidationError):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

    def test_create_repo_handle_get_error(self):
        """
        Test GitLabProvider handles GitLab repo GET errors
        """
        self.gl_mock().projects.get.side_effect = GitlabGetError

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'

        with self.assertRaises(APIException):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)

    def test_create_repo_handle_hook_create_error(self):
        """
        Test GitLabProvider handles GitLab hook creation errors
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'
        self.gl_mock().projects.get.return_value.hooks.create.side_effect = GitlabCreateError

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'

        with self.assertRaises(APIException):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().hooks.create.call_count, 1)

    @patch('arkindex.dataimport.providers.git')
    def test_clone_repo(self, git_mock):
        """
        Test GitLabProvider can clone a repository
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).clone_repo(self.repo, 'somewhere', a='a', b='b')

        self.assertEqual(git_mock.Repo.clone_from.call_count, 1)
        args, kwargs = git_mock.Repo.clone_from.call_args
        self.assertTupleEqual(args, ('http://oauth2:oauth-token@gitlab/repo', 'somewhere'))
        self.assertDictEqual(kwargs, {'a': 'a', 'b': 'b'})

    @patch('arkindex.dataimport.providers.open')
    def test_download_archive(self, open_mock):
        """
        Test GitLabProvider can download an archive of a revision onto a specified path
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).download_archive(self.rev, 'somewhere')

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        args, kwargs = self.gl_mock().projects.get.call_args
        self.assertTupleEqual(args, ('repo', ))
        self.assertDictEqual(kwargs, {})

        self.assertEqual(open_mock.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().repository_archive.call_count, 1)
        args, kwargs = self.gl_mock().projects.get().repository_archive.call_args
        self.assertTupleEqual(args, ())
        self.assertDictEqual(kwargs, {
            'sha': '42',
            'streamed': True,
            'action': open_mock().__enter__().write,
        })

    @patch('arkindex.dataimport.providers.open')
    def test_download_archive_get_error(self, open_mock):
        """
        Test GitLabProvider can handle GitLab errors while downloading an archive
        """
        self.gl_mock().projects.get.side_effect = GitlabGetError

        with self.assertRaises(APIException):
            GitLabProvider(url='http://aaa', credentials=self.creds).download_archive(self.rev, 'somewhere')

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(open_mock.call_count, 0)

    def test_get_revision(self):
        """
        Test GitLabProvider can create a Revision instance for a repo by hash
        """
        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_revision(self.repo, '42')

        self.assertEqual(revision, self.rev)
        self.assertFalse(created)
        self.assertEqual(self.gl_mock.call_count, 0)

    def test_create_revision(self):
        """
        Test GitLabProvider can create a Revision instance for a repo by hash
        """
        self.gl_mock().projects.get.return_value.commits.get.return_value.refs.return_value = [
            {'name': 'oh_no', 'type': 'tag'},
            {'name': 'refs/heads/branch1', 'type': 'branch'},
            {'name': 'refs/heads/branch2', 'type': 'branch'},
        ]
        self.gl_mock().projects.get.return_value.commits.get.return_value.message = 'commit message'
        self.gl_mock().projects.get.return_value.commits.get.return_value.author_name = 'bob'

        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_revision(self.repo, '1337')

        self.assertTrue(created)
        self.assertEqual(revision.hash, '1337')
        self.assertEqual(revision.ref, 'refs/heads/branch1')
        self.assertEqual(revision.message, 'commit message')
        self.assertEqual(revision.author, 'bob')

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().commits.get.call_count, 1)
        args, kwargs = self.gl_mock().projects.get().commits.get.call_args
        self.assertTupleEqual(args, ('1337', ))
        self.assertDictEqual(kwargs, {})

    def test_get_latest_revision(self):
        """
        Test GitLabProvider can get the latest revision on a repo
        """
        latest_commit = MagicMock()
        latest_commit.id = '42'
        latest_commit.refs.return_value = [
            {'name': 'refs/heads/master'},
        ]
        latest_commit.message = 'a'
        latest_commit.author_name = 'me'

        self.gl_mock().projects.get.return_value.commits.list.return_value = [latest_commit, ]

        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_latest_revision(self.repo)

        self.assertFalse(created)
        self.assertEqual(revision, self.rev)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().commits.list.call_count, 1)

    def test_create_latest_revision(self):
        """
        Test GitLabProvider can create the latest revision on a repo
        """
        latest_commit = MagicMock()
        latest_commit.id = '1337'
        latest_commit.refs.return_value = [
            {'name': 'refs/heads/branch1'},
            {'name': 'refs/heads/branch2'},
        ]
        latest_commit.message = 'commit message'
        latest_commit.author_name = 'bob'

        self.gl_mock().projects.get.return_value.commits.list.return_value = [latest_commit, ]

        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_latest_revision(self.repo)

        self.assertTrue(created)
        self.assertEqual(revision.hash, '1337')
        self.assertEqual(revision.ref, 'refs/heads/branch1')
        self.assertEqual(revision.message, 'commit message')
        self.assertEqual(revision.author, 'bob')

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().commits.list.call_count, 1)

    def test_handle_webhook(self):
        """
        Test GitLabProvider correctly handles GitLab webhook push events
        """
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': '1337',
            'commits': [
                {
                    'message': 'commit message',
                    'author': {
                        'name': 'bob',
                    }
                }
            ]
        }

        repo_imports = DataImport.objects.filter(payload__repo_id=str(self.repo.id))
        self.assertFalse(repo_imports.exists())
        GitLabProvider(url='http://aaa', credentials=self.creds).handle_webhook(self.repo, request_mock)
        di = repo_imports.get()
        self.assertEqual(di.mode, DataImportMode.Repository)
        self.assertEqual(di.workflow.state, State.Unscheduled)

    def test_handle_webhook_missing_headers(self):
        """
        Test GitLabProvider checks HTTP headers on webhooks
        """
        glp = GitLabProvider(url='http://aaa', credentials=self.creds)

        request_mock = MagicMock()
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': '1337',
            'commits': [
                {
                    'message': 'commit message',
                    'author': {
                        'name': 'bob',
                    }
                }
            ]
        }

        # Missing HTTP_X_GITLAB_TOKEN
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
        }
        with self.assertRaises(NotAuthenticated):
            glp.handle_webhook(self.repo, request_mock)

        # Missing HTTP_X_GITLAB_EVENT
        request_mock.META = {
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)

        # Wrong HTTP_X_GITLAB_EVENT
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Not a Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)

        # Wrong HTTP_X_GITLAB_TOKEN
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'not-the-hook-token',
        }
        with self.assertRaises(AuthenticationFailed):
            glp.handle_webhook(self.repo, request_mock)

    def test_handle_webhook_duplicate_events(self):
        """
        Test GitLabProvider checks for already handled events
        """
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': '42',
            'commits': [
                {
                    'message': 'a',
                    'author': {
                        'name': 'me',
                    }
                }
            ]
        }

        repo_imports = DataImport.objects.filter(payload__repo_id=str(self.repo.id))
        self.assertFalse(repo_imports.exists())
        GitLabProvider(url='http://aaa', credentials=self.creds).handle_webhook(self.repo, request_mock)
        self.assertFalse(repo_imports.exists())

    def test_handle_webhook_wrong_kind(self):
        """
        Test GitLabProvider gracefully fails when GitLab breaks things
        """
        repo_imports = DataImport.objects.filter(payload__repo_id=str(self.repo.id))
        glp = GitLabProvider(url='http://aaa', credentials=self.creds)
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        request_mock.data = {
            'object_kind': 'oh_no',  # Bad object kind
            'ref': 'refs/heads/master',
            'checkout_sha': '1337',
            'commits': [
                {
                    'message': 'b',
                    'author': {
                        'name': 'me',
                    }
                }
            ]
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(repo_imports.exists())

        # Missing SHA
        request_mock.data['object_kind'] = 'push'
        del request_mock.data['checkout_sha']
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(repo_imports.exists())

        # Breaking change: a list!
        request_mock.data = [request_mock.data]
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(repo_imports.exists())
