from django.db.models import Prefetch, prefetch_related_objects
from django_filters import rest_framework as filters
from rest_framework.exceptions import ValidationError
from rest_framework.generics import (
    ListAPIView, CreateAPIView, RetrieveAPIView, UpdateAPIView, DestroyAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView
)
from rest_framework import status, response
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Classification, Corpus, Element, ElementPath, Right, Transcription, Region
from arkindex.documents.serializers.elements import (
    CorpusSerializer, ElementSerializer, ElementSlimSerializer,
    ElementCreateSerializer, ElementNeighborsSerializer, ElementParentSerializer,
    RegionSerializer, RegionCreateSerializer, RegionBulkSerializer,
)
from arkindex.project.openapi import AutoSchema
from arkindex.documents.serializers.light import RegionLightSerializer, MetaDataLightSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly

from .filters import ElementsMLClassFilter
from uuid import UUID


classifications_prefetch = Prefetch(
    'classifications',
    queryset=Classification.objects.select_related('ml_class', 'source').order_by('-confidence')
)


class ElementsList(CorpusACLMixin, ListAPIView):
    """
    List elements and filter by type, name, corpus, ML class
    """
    serializer_class = ElementSlimSerializer
    queryset = Element.objects.all()
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = ElementsMLClassFilter
    openapi_overrides = {
        'operationId': 'ListElements',
        'description': 'List elements and filter by type, name, corpus, ML class',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'corpus',
                'in': 'query',
                'description': 'Filter elements by corpus ID',
                'required': False,
                'schema': {'type': 'string', 'format': 'uuid'},
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter by element type',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'hidden',
                'in': 'query',
                'description': 'Restrict to or exclude elements with hidden types; they are excluded by default',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
        ],
    }

    def filter_queryset(self, queryset):
        filters = {}
        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        if 'type' in self.request.query_params:
            filters['type__slug'] = self.request.query_params['type']
        else:
            # Hide hidden elements if there isn't a type filter
            filters['type__hidden'] = False

        only_hidden = self.request.query_params.get('hidden')
        if only_hidden is not None:
            filters['type__hidden'] = only_hidden.lower() not in ('false', '0')

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        if 'corpus' in self.request.query_params:
            try:
                corpus_id = UUID(self.request.query_params['corpus'])
            except ValueError:
                raise ValidationError({'corpus': ['Not a valid uuid']})
            filters['corpus'] = self.get_corpus(corpus_id)
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)

        return super().filter_queryset(queryset) \
            .filter(**filters) \
            .prefetch_related('corpus', 'zone__image__server', 'type') \
            .order_by('corpus', 'type__slug', 'name')


class ElementRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        if self.request.method == 'GET':
            corpora = Corpus.objects.readable(self.request.user)
        else:
            corpora = Corpus.objects.writable(self.request.user)
        return Element.objects \
            .filter(corpus__in=corpora) \
            .select_related(
                'corpus',
                'type',
                'zone__image__server',
            ) \
            .prefetch_related(
                'metadatas__revision__repo',
                'metadatas__entity',
                classifications_prefetch,
            )

    def perform_destroy(self, instance):
        children_count = ElementPath.objects.filter(path__contains=[instance.id]).count()
        if children_count:
            raise ValidationError({
                'element': "Element '{}' is linked to {} children elements".format(instance.id, children_count)
            })
        super().perform_destroy(instance)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET':
            context['element'] = self.get_object()
        return context


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementSlimSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'

    def get_queryset(self):
        filtering = {
            'corpus__in': Corpus.objects.readable(self.request.user),
        }
        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filtering['type'] = type_param
        related = Element.objects.get_related(self.kwargs['pk'], **filtering)
        prefetch_related_objects(related, 'corpus', 'zone__image__server')
        return related


class ElementNeighbors(ListAPIView):
    """
    List neighboring elements
    """
    serializer_class = ElementNeighborsSerializer

    def get_queryset(self):

        def presentation(n):
            # Preserve the ordering defined in the path: highest-level parents come first
            parents = sorted(
                Element.objects.filter(id__in=n.path),
                key=lambda element: n.path.index(element.id),
            )
            return {'position': n.ordering, 'element': n.element, 'parents': parents}

        n = int(self.request.query_params.get('n', 1))
        element = Element.objects.get(id=self.kwargs['pk'])
        neighbors = Element.objects.get_neighbors(element, n)
        return [presentation(n) for n in neighbors]


class NoRequestBodySchema(AutoSchema):
    def _get_request_body(self, path, method):
        # Remove request body from OpenAPI schema
        return None


class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )
    schema = NoRequestBodySchema()

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return response.Response(status=status.HTTP_204_NO_CONTENT)


class ElementParents(ListAPIView):
    """
    List all parents of an element
    """
    serializer_class = ElementSlimSerializer
    openapi_overrides = {
        'operationId': 'ListElementParents',
        'description': 'List all parents of an element',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter elements by type',
                'required': False,
                'schema': {
                    'type': 'string',
                }
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (children, grandchildren, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
            {
                'name': 'hidden',
                'in': 'query',
                'description': 'Restrict to or exclude elements with hidden types; they are excluded by default',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
        ]
    }

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user)
        }

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filters['type__slug'] = type_param

        only_hidden = self.request.query_params.get('hidden')
        if only_hidden is not None:
            filters['type__hidden'] = only_hidden.lower() not in ('false', '0')

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        recursive_param = self.request.query_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # List direct parents only: elements whose IDs are the last in the element's paths
            return Element.objects.filter(
                id__in=ElementPath.objects
                                  .filter(element_id=self.kwargs['pk'])
                                  .values('path__last'),
                **filters,
            ).prefetch_related('zone__image__server', 'corpus', 'type')

        parents = Element.objects.get_ascending(self.kwargs['pk'], **filters)
        prefetch_related_objects(parents, 'zone__image__server', 'corpus', 'type')
        return parents


class ElementChildren(ListAPIView):
    """
    List all children of an element
    """
    serializer_class = ElementSlimSerializer
    openapi_overrides = {
        'operationId': 'ListElementChildren',
        'description': 'List all children of an element',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter elements by type',
                'required': False,
                'schema': {
                    'type': 'string',
                }
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (parents, grandparents, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
            {
                'name': 'hidden',
                'in': 'query',
                'description': 'Restrict to or exclude elements with hidden types; they are excluded by default',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
        ]
    }

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user)
        }

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filters['type__slug'] = type_param

        only_hidden = self.request.query_params.get('hidden')
        if only_hidden is not None:
            filters['type__hidden'] = only_hidden.lower() not in ('false', '0')

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        recursive_param = self.request.query_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # Only list direct children
            # We could use Element.objects.filter(paths__path__last=volid) directly,
            # but it is 100× slower due to the implementation of path__last.
            # get_descending uses 'path @> ARRAY[element.id]', which is much faster,
            # so we append the __last filter to it.
            filters['paths__path__last'] = self.kwargs['pk']

        return Element.objects \
            .get_descending(self.kwargs['pk']) \
            .filter(**filters) \
            .prefetch_related('zone__image__server', 'corpus', 'type')


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user).prefetch_related('types')


class CorpusRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method != 'GET' and Right.Write not in obj.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    Get transcriptions corresponding to a specific element
    Transcriptions can be filtered by type
    """
    serializer_class = TranscriptionSerializer
    pagination_class = TranscriptionsPagination

    def get_queryset(self):
        queryset = Transcription.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        ).prefetch_related('zone__image__server', 'source').order_by('id')
        req_type = self.request.query_params.get('type')
        if req_type:
            try:
                req_type = TranscriptionType(req_type)
                queryset = queryset.filter(type=req_type)
            except ValueError:
                raise ValidationError({'type': 'Not a valid transcription type'})
        return queryset


class ElementRegions(ListAPIView):
    serializer_class = RegionLightSerializer

    def get_queryset(self):
        return Region.objects \
            .filter(
                elements=self.kwargs['pk'],
                elements__corpus__in=Corpus.objects.readable(self.request.user),
            ) \
            .distinct() \
            .select_related('ml_class', 'source') \
            .prefetch_related('zone__image__server') \
            .order_by('region_elements__ordering')


class ElementsCreate(CreateAPIView):
    """
    Add a child to an existing element
    Create a new Element
    A parent and/or an image can be attached to this Element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer


class RegionDetails(RetrieveAPIView):
    serializer_class = RegionSerializer

    def get_queryset(self):
        return Region.objects \
            .filter(elements__corpus__in=Corpus.objects.readable(self.request.user)) \
            .select_related('ml_class', 'source') \
            .prefetch_related('zone__image__server', 'elements__type')


class RegionCreate(CreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RegionCreateSerializer

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context


class RegionBulkCreate(CreateAPIView, UpdateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RegionBulkSerializer
    # Force DRF to ignore PATCH
    http_method_names = ['post', 'put', 'head', 'options', 'trace']
    # Cannot override operationId or description here because openapi_overrides does not handle multiple HTTP methods
    openapi_overrides = {
        'tags': ['ml'],
    }

    def get_object(self):  # Ignore the get_object part in UpdateAPIView.update
        return

    def perform_create(self, serializer):
        serializer.create(serializer.validated_data)

    def perform_update(self, serializer):
        serializer.update(None, serializer.validated_data)


class ElementMetadata(CreateAPIView):
    """
    Create a medadata on an existing element
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataLightSerializer
    openapi_overrides = {
        'operationId': 'CreateMetaData',
        'tags': ['elements']
    }

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context
