from django.db.models import Q, Count, F, Window
from django.db.models.functions import FirstValue
from django_filters import rest_framework as filters
from arkindex.documents.models import MLClass, Element, ClassificationState, Classification


class MLClassFolderChoiceFilter(filters.ModelChoiceFilter):
    def get_queryset(self, request):
        return request.corpus.elements.filter(type__folder=True)


class CorpusMLClassFilter(filters.FilterSet):
    folder = MLClassFolderChoiceFilter(method='filter_folder', label='Restrict classes to elements in a folder')

    def filter_folder(self, queryset, name, value):
        return queryset.filter(classifications__element__in=Element.objects.get_descending(value.pk)).distinct()

    class Meta:
        model = MLClass
        fields = ['folder']


class ElementsMLClassFilter(filters.FilterSet):
    ml_class = filters.ModelChoiceFilter(
        queryset=MLClass.objects.all(),
        field_name='classifications__ml_class',
        label='Class',
        method='filter_ml_class'
    )
    ml_class_unvalidated = filters.BooleanFilter(
        field_name='classifications__ml_class', method='filter_ml_class_unvalidated')

    def filter_ml_class_unvalidated(self, queryset, name, value):
        base_qs = queryset.annotate(
            num_validated=Count('classifications', filter=Q(classifications__state=ClassificationState.Validated)))
        if value:
            return base_qs.filter(num_validated=0)
        return base_qs.exclude(num_validated=0)

    def filter_ml_class(self, queryset, name, value):
        element_ids = Window(
            FirstValue('id'),
            partition_by=[F('element_id')],
            order_by=F('confidence').desc()
        )

        return queryset.filter(
            classifications__ml_class_id=value.id,
            classifications__id__in=Classification.objects.annotate(window=element_ids).values('window')
        )

    class Meta:
        model = Element
        fields = ['ml_class', 'ml_class_unvalidated']
