#!/usr/bin/env python3
from django.conf import settings
from django.core.management.base import CommandError
from ponos.management.base import PonosCommand
from arkindex.project.argparse import CorpusArgument, ElementArgument
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Element, Transcription, Entity
import logging

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)
logger = logging.getLogger(__name__)


def get_transcriptions(corpus=None, folder=None):
    if folder:
        # Lookup all the transcriptions linked to a folder
        return Transcription.objects.filter(
            element__in=Element.objects.get_descending(folder.id)
        ).distinct()
    elif corpus:
        return Transcription.objects.filter(element__corpus=corpus)

    return Transcription.objects.all()


def get_elements(corpus=None, folder=None):
    if folder:
        queryset = Element.objects.get_descending(folder.id)
    elif corpus:
        queryset = Element.objects.filter(corpus=corpus)
    else:
        queryset = Element.objects.all()

    return queryset \
        .filter(type__folder=False, type__hidden=False) \
        .prefetch_related('metadatas', 'transcriptions')


def get_entities(corpus=None, folder=None):
    if folder:
        return Entity.objects.filter(
            metadatas__element__in=Element.objects.get_descending(folder.id)
        ).distinct()
    elif corpus:
        return Entity.objects.filter(corpus=corpus)

    return Entity.objects.all()


class Command(PonosCommand):
    help = 'Selectively reindex transcriptions, elements and entities into ElasticSearch'
    docker_image = settings.ARKINDEX_APP_IMAGE
    base_recipe = settings.PONOS_RECIPE

    # Setup for reindexation of different Elements
    index_methods = {
        'elements': {
            'bulk_size': 100,
            'model': Element,
            'items': get_elements,
        },
        'transcriptions': {
            'bulk_size': 400,
            'model': Transcription,
            'items': get_transcriptions,
        },
        'entities': {
            'bulk_size': 400,
            'model': Entity,
            'items': get_entities,
        }
    }

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '-ts', '--transcriptions',
            help='Reindex transcriptions',
            action='store_true',
        )
        parser.add_argument(
            '--elements',
            help='Reindex elements',
            action='store_true',
        )
        parser.add_argument(
            '--entities',
            help='Reindex entities',
            action='store_true',
        )
        parser.add_argument(
            '--corpus',
            help='Restrict reindexing to a specific corpus by ID or part of the name',
            type=CorpusArgument(),
        )
        parser.add_argument(
            '--folder',
            help='Restrict reindexing to a specific folder by ID or part of the name',
            type=ElementArgument(type__folder=True),
        )
        parser.add_argument(
            '--drop',
            help="Drop the existing indexes before reindexing",
            action='store_true',
        )

    def validate_args(self, folder=None, corpus=None, **options):
        if corpus and folder and folder.corpus != corpus:
            raise CommandError('Folder {} is not in corpus {}'.format(folder, corpus))

        # If no specific index was set, reindex everything
        if not any(options.get(k) for k in self.index_methods.keys()):
            for k in self.index_methods.keys():
                options[k] = True

        return {
            'methods': [
                key
                for key in self.index_methods.keys()
                if options.get(key)
            ],
            'corpus': corpus,
            'folder': folder,
            'drop': options.get('drop', False),
        }

    def run(self, methods=[], drop=False, corpus=None, folder=None):
        '''
        Index all prepared elements
        '''
        indexer = Indexer()
        if drop:
            for method in methods:
                config = self.index_methods[method]
                indexer.drop_index(config['model'].es_document)
            indexer.setup()

        for method in methods:
            config = self.index_methods[method]

            items = config['items'](corpus=corpus, folder=folder)

            indexer.run_index(
                items,
                bulk_size=config['bulk_size'],
            )
