# Generated by Django 2.2 on 2019-09-17 08:20

from django.db import migrations
from django.db.models import Prefetch
from arkindex_common.enums import MetaType


def fields_to_metadata(apps, schema_editor):
    """
    Translate Page.folio, Act.folio and Act.number fields into metadatas
    on their elements before the subclasses get removed
    """
    Page = apps.get_model('documents', 'page')
    Act = apps.get_model('documents', 'act')
    MetaData = apps.get_model('documents', 'MetaData')

    def _field_to_metadata(model, name):
        queryset = model.objects.prefetch_related(Prefetch(
            'metadatas',
            queryset=MetaData.objects.filter(name=name),
        ))
        if not queryset.exists():
            return
        print('Migrating {!r} on {} {}s'.format(name, queryset.count(), model.__name__))
        metadatas = []
        for element in queryset:
            index = 0
            if element.metadatas.exists():
                index = max(md.index for md in element.metadatas.all() if md.name == name) + 1
            value = getattr(element, name)
            if not value:  # Ignore empty metadatas
                continue
            metadatas.append(MetaData(
                element_id=element.id,
                name=name,
                type=MetaType.Text,
                value=value,
                index=index,
            ))

        print('Creating {} metadatas'.format(len(metadatas)))
        MetaData.objects.bulk_create(metadatas)

    _field_to_metadata(Page, 'folio')
    _field_to_metadata(Act, 'folio')
    _field_to_metadata(Act, 'number')


def metadata_to_fields(apps, schema_editor):
    """
    Rollback: Find pages and acts as *elements*, get their folio/number metadata, and
    rebuild the Page and Act instances from existing Element instances, which is not
    something Django handles well.
    """
    Element = apps.get_model('documents', 'Element')
    Page = apps.get_model('documents', 'Page')
    Act = apps.get_model('documents', 'Act')
    MetaData = apps.get_model('documents', 'MetaData')

    if not Element.objects.filter(type__in=('page', 'act')).exists():  # Empty database
        return

    page_elements = Element.objects.filter(
        type='page',
        metadatas__name='folio',
    ).distinct().prefetch_related(Prefetch(
        'metadatas',
        queryset=MetaData.objects.filter(name='folio'),
        to_attr='folio_metadatas',
    ))
    print('Rolling back {} pages'.format(page_elements.count()))

    metadata_ids, pages = [], []
    for element in page_elements:
        assert element.folio_metadatas, '{} ({}) has no folio metadatas'.format(element.name, element.id)
        if len(element.folio_metadatas) > 1:
            print('{} ({}) has more than one folio metadata; picking the first one'.format(element.name, element.id))
        folio_metadata = element.folio_metadatas[0]
        metadata_ids.append(folio_metadata.id)

        pages.append(Page(
            id=element.id,
            element_ptr_id=element.id,
            folio=folio_metadata.value,
        ))

    # Cannot use bulk_create because it is unsupported on a multi-table model
    # Call the underlying _batched_insert instead
    # https://github.com/django/django/blob/b1d6b35e146aea83b171c1b921178bbaae2795ed/django/db/models/query.py#L480
    print('Saving {} pages'.format(len(pages)))
    Page.objects.get_queryset()._batched_insert(
        pages,
        fields=Page._meta.local_fields[:2],  # Field instances: element_ptr_id and folio
        batch_size=None,
    )

    print('Removing {} metadatas'.format(len(metadata_ids)))
    MetaData.objects.filter(id__in=metadata_ids).delete()

    act_elements = Element.objects.filter(
        type='act',
        metadatas__name__in=('folio', 'number'),
    ).distinct().prefetch_related(
        Prefetch(
            'metadatas',
            queryset=MetaData.objects.filter(name='folio'),
            to_attr='folio_metadatas',
        ),
        Prefetch(
            'metadatas',
            queryset=MetaData.objects.filter(name='number'),
            to_attr='number_metadatas',
        ),
    )
    print('Rolling back {} acts'.format(act_elements.count()))

    metadata_ids, acts = [], []
    for element in act_elements:
        assert element.folio_metadatas, '{} ({}) has no folio metadatas'.format(element.name, element.id)
        if len(element.folio_metadatas) > 1:
            print('{} ({}) has more than one folio metadata; picking the first one'.format(element.name, element.id))
        folio_metadata = element.folio_metadatas[0]

        assert element.number_metadatas, '{} ({}) has no number metadatas'.format(element.name, element.id)
        if len(element.number_metadatas) > 1:
            print('{} ({}) has more than one number metadata; picking the first one'.format(element.name, element.id))
        number_metadata = element.number_metadatas[0]

        metadata_ids.append(folio_metadata.id)
        metadata_ids.append(number_metadata.id)
        acts.append(Act(
            id=element.id,
            element_ptr_id=element.id,
            folio=folio_metadata.value,
            number=number_metadata.value,
        ))

    # Cannot use bulk_create because it is unsupported on a multi-table model
    print('Saving {} acts'.format(len(acts)))
    Act.objects.get_queryset()._batched_insert(
        acts,
        fields=Act._meta.local_fields,  # Field instances: element_ptr_id, number, folio
        batch_size=None,
    )

    print('Removing {} metadatas'.format(len(metadata_ids)))
    MetaData.objects.filter(id__in=metadata_ids).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('documents', '0016_classification_on_element'),
    ]

    operations = [
        migrations.RunPython(fields_to_metadata, metadata_to_fields, elidable=True),
    ]
