from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import MetaType
from arkindex.documents.models import \
    Corpus, Element, ElementPath, ElementType, MetaData, Region, RegionElement, DataSource, MLClass
from arkindex.images.serializers import ZoneSerializer
from arkindex.images.models import Image
from arkindex.documents.serializers.light import \
    CorpusLightSerializer, ElementLightSerializer, ElementTypeSerializer, \
    EntityLightSerializer, RegionLightSerializer, MetaDataLightSerializer
from arkindex.documents.serializers.ml import ClassificationSerializer
from arkindex.dataimport.models import EventType
from arkindex.project.serializer_fields import PolygonField, MLToolField
from arkindex.project.polygon import Polygon
from arkindex.project.aws import S3FileStatus
from collections import defaultdict
import uuid


class MetaDataSerializer(MetaDataLightSerializer):
    """
    Serialises some Metadata for any Element
    """
    entity = EntityLightSerializer()

    class Meta:
        model = MetaData
        fields = MetaDataLightSerializer.Meta.fields + ('entity', )


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """
    rights = serializers.SerializerMethodField(read_only=True)
    types = ElementTypeSerializer(many=True, read_only=True)

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'description',
            'public',
            'rights',
            'types',
        )

    def get_rights(self, corpus):
        return map(lambda x: x.value, corpus.get_acl_rights(self.context['request'].user))

    def create(self, validated_data):
        corpus = Corpus.objects.create(**validated_data)
        corpus.corpus_right.create(
            user=self.context['request'].user,
            can_write=True,
            can_admin=True,
        )
        return corpus


class ElementSlimSerializer(serializers.ModelSerializer):
    """
    Fully Serialises a document
    """
    type = serializers.SlugRelatedField(queryset=ElementType.objects.none(), slug_field='slug')
    corpus = CorpusLightSerializer(read_only=True)
    thumbnail_url = serializers.SerializerMethodField(read_only=True)
    thumbnail_put_url = serializers.SerializerMethodField(read_only=True)
    zone = ZoneSerializer(read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('element'):
            # Do not raise Error in order to create OpenAPI schema
            return
        self.fields['type'].queryset = ElementType.objects.filter(corpus=self.context['element'].corpus)

    def validate_type(self, new_type):
        elt_type = self.context['element'].type
        if (elt_type.folder, elt_type.hidden) != (new_type.folder, new_type.hidden):
            raise ValidationError(
                "'{}' type does not have the same properties as '{}' type"
                .format(new_type.slug, self.context['element'].type.slug)
            )
        return new_type

    def get_thumbnail_url(self, element):
        if element.type.folder:
            return element.thumbnail.s3_url

    def get_thumbnail_put_url(self, element):
        user = self.context['request'].user
        if user.is_authenticated and (user.is_admin or user.is_internal) and element.type.folder:
            return element.thumbnail.s3_put_url

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'corpus',
            'thumbnail_url',
            'thumbnail_put_url',
            'zone',
        )
        read_only_fields = (
            'id',
            'corpus',
            'thumbnail_url',
            'thumbnail_put_url',
            'zone',
        )


class ElementParentSerializer(serializers.Serializer):
    """
    Serialize a relation between an element and one of its parents
    """
    child = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())
    parent = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['child'].queryset = Element.objects.filter(corpus__in=corpora)
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora)

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)
        child = data.get('child')
        parent = data.get('parent')

        errors = defaultdict(list)
        if parent.corpus != child.corpus:
            errors['parent'].append("Parent is not from corpus '{}'".format(child.corpus.name))
        if parent.id == child.id:
            errors['parent'].append('A child cannot be its own parent')
        if parent.type.folder and child.type.hidden:
            errors['parent'].append(
                "Element '{}' has folder property thus cannot contain hidden elements".format(parent.id)
            )
        # A non folder parent could have hidden childs
        if not parent.type.folder and not child.type.hidden:
            errors['parent'].append(
                "Element '{}' does not have folder property and may "
                "contain hidden elements only".format(parent.id)
            )
        if errors:
            raise ValidationError(errors)
        return data

    def perform_create_checks(self, child, parent):
        # Assert parent is not an child's ancestor already
        if ElementPath.objects.filter(element_id=child.id, path__contains=[parent.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is already a parent of '{}'".format(parent.id, child.id)
            ]})
        # Assert parent is not an alement's descendant
        if ElementPath.objects.filter(element_id=parent.id, path__contains=[child.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is a child of element '{}'".format(parent.id, child.id)
            ]})

    def create(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        self.perform_create_checks(child, parent)
        child.add_parent(parent)

    def delete(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        parent.remove_child(child)


class ElementSerializer(ElementSlimSerializer):
    """
    Serialize an element with its metadata and classifications
    """

    metadata = serializers.SerializerMethodField()
    classifications = ClassificationSerializer(many=True, read_only=True)

    class Meta:
        model = Element
        fields = ElementSlimSerializer.Meta.fields + ('metadata', 'classifications')
        read_only_fields = ElementSlimSerializer.Meta.read_only_fields + ('metadata', 'classifications')

    def create(self, *args, **kwargs):
        instance = super().create(*args, **kwargs)
        instance.events.create(type=EventType.Addition)
        return instance

    def update(self, *args, **kwargs):
        instance = super().update(*args, **kwargs)
        instance.events.create(type=EventType.Edit)
        return instance

    def get_metadata(self, obj):
        return MetaDataLightSerializer(
            list(obj.metadatas.all().filter(entity__isnull=True).exclude(type=MetaType.Entity)),
            many=True
        ).data


class ElementNeighborsSerializer(serializers.Serializer):
    position = serializers.IntegerField()
    element = ElementLightSerializer()
    parents = serializers.ListField(
        child=ElementLightSerializer()
    )


class ElementCreateSerializer(ElementLightSerializer):
    """
    Serialize an Element with a possible parent and image
    """
    type = serializers.SlugField(source='type.slug')
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.all().prefetch_related('server'),
        required=False
    )
    parent = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none(), required=False)

    class Meta(ElementLightSerializer.Meta):
        model = Element
        fields = ElementLightSerializer.Meta.fields + ('image', 'corpus', 'parent')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora)

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)

        try:
            data['type'] = data['corpus'].types.get(slug=data['type']['slug'])
        except ElementType.DoesNotExist:
            # Fail immediately here because the type is required for other checks
            raise ValidationError({
                'type': ['Type {!r} not found in corpus {}'.format(data['type']['slug'], data['corpus'].name)]
            })

        parent = data.get('parent')
        if parent and parent.corpus != data['corpus']:
            errors['corpus'].append(
                'Parent and child must be in the same corpus'
            )
        image = data.get('image')
        if image and image.status != S3FileStatus.Checked:
            errors['image'].append(
                'Image is not checked. Try to upload a valid IIIF image.'
            )

        # Limit Elements creation cases
        if parent:
            if data['type'].hidden:
                if parent.type.folder:
                    errors['parent'].append(
                        'The parent of a hidden element should not be a folder'
                    )
                if parent.type.hidden:
                    errors['parent'].append(
                        'The parent of a hidden element should not be hidden'
                    )
            elif not parent.type.folder:
                errors['parent'].append(
                    'The parent of a folder or a normal element should be a folder'
                )

        elif not data['type'].folder:
            errors['parent'].append(
                'Non-folder elements cannot be created without a parent folder'
            )

        if errors:
            raise ValidationError(errors)
        return data

    def create(self, validated_data):
        image = validated_data.pop('image', None)
        parent = validated_data.pop('parent', None)
        metadata = validated_data.pop('metadata', {})

        element = Element(**validated_data)
        element.full_clean()

        if image:
            # Create a new zone containing the image
            zone = image.zones.create(polygon=Polygon.from_coords(0, 0, image.width, image.height))
            # update Element zone
            element.zone = zone

        element.save()
        # Make relation to parent
        if parent:
            element.add_parent(parent)

        for name, value in metadata.items():
            element.metadatas.create(name=name, type=MetaType.Text, value=value)

        return element


class RegionSerializer(RegionLightSerializer):
    elements = ElementLightSerializer(many=True)

    class Meta(RegionLightSerializer.Meta):
        fields = RegionLightSerializer.Meta.fields + ('elements', )


class RegionCreateSerializer(serializers.Serializer):
    id = serializers.UUIDField(read_only=True)
    class_name = serializers.CharField(source='ml_class')
    confidence = serializers.FloatField(min_value=0, max_value=1)
    polygon = PolygonField()
    analyser = MLToolField(MLToolType.DLAnalyser)

    def validate_class_name(self, name):
        assert 'element' in self.context, 'Missing associated element'
        return self.context['element'].corpus.ml_classes.get_or_create(name=name)[0]

    def save(self):
        polygon = self.validated_data['polygon']

        # Do not use get_or_create due to possible duplicate zones
        zone = self.context['element'].zone.image.zones.filter(polygon=polygon).first()
        if not zone:
            zone = self.context['element'].zone.image.zones.create(polygon=polygon)

        region, _ = Region.objects.get_or_create(
            ml_class=self.validated_data['ml_class'],
            zone=zone,
            confidence=self.validated_data['confidence'],
            source=DataSource.from_ml_tool(self.validated_data['analyser']),
        )
        region.elements.add(self.context['element'])
        self.validated_data['id'] = region.id
        return region


class RegionBulkItemSerializer(serializers.Serializer):
    polygon = PolygonField()
    class_name = serializers.CharField()
    confidence = serializers.FloatField(min_value=0, max_value=1)


class RegionBulkSerializer(serializers.Serializer):
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.all(),
    )
    analyser = MLToolField(MLToolType.DLAnalyser, source='source')
    regions = RegionBulkItemSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )

    def validate_analyser(self, value):
        return DataSource.from_ml_tool(value)

    def validate(self, data):
        data = super().validate(data)
        corpus_id = data['parent'].corpus_id
        # Retrieve or create required ML classes
        class_names = [region['class_name'] for region in data['regions']]
        ml_classes = {
            class_name: MLClass.objects.get_or_create(
                corpus_id=corpus_id,
                name=class_name,
            )[0]
            for class_name in class_names
        }
        # Add ML classes to regions
        for region in data['regions']:
            region['ml_class'] = ml_classes[region['class_name']]
        return data

    def update(self, instance, validated_data):
        # Remove previous regions from the same source on this element
        validated_data['parent'].regions.filter(source=validated_data['source']).delete()
        self.create(validated_data)

    def create(self, validated_data):
        element = validated_data['parent']
        image = element.zone.image
        source = validated_data['source']

        regions, region_elements = [], []
        for region_data in validated_data['regions']:
            # TODO: Replace with get_or_create once zones are deduplicated
            zone = image.zones.filter(polygon=region_data['polygon']).first()
            if not zone:
                zone = image.zones.create(polygon=region_data['polygon'])

            region = Region(
                id=uuid.uuid4(),
                zone=zone,
                source=source,
                ml_class=region_data['ml_class'],
                confidence=region_data['confidence'],
            )
            regions.append(region)
            region_elements.append(RegionElement(
                region_id=region.id,
                element_id=element.id,
            ))

        Region.objects.bulk_create(regions)
        RegionElement.objects.bulk_create(region_elements)
        return regions
