from rest_framework import serializers
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import TranscriptionType
from arkindex.project.serializer_fields import EnumField, MLToolField, PolygonField
from arkindex.documents.models import (
    Corpus, Element, Transcription, DataSource, MLClass, Classification, ClassificationState
)
from arkindex.images.serializers import ZoneSerializer


class DataSourceSerializer(serializers.ModelSerializer):
    """
    Serialize a data source for transcriptions and classifications
    """

    type = EnumField(MLToolType)

    class Meta:
        model = DataSource
        fields = (
            'id',
            'type',
            'slug',
            'name',
            'revision',
            'internal',
        )


class MLClassSerializer(serializers.ModelSerializer):
    """
    Serializer for MLClass instances
    Used as a nested endpoint below corpus, so exclude corpus from payload
    """
    class Meta:
        model = MLClass
        exclude = ('corpus',)


class ClassificationSerializer(serializers.ModelSerializer):
    """
    Serialize a classification on an Element
    """

    source = DataSourceSerializer()
    ml_class = MLClassSerializer()
    state = EnumField(ClassificationState)

    class Meta:
        model = Classification
        fields = (
            'id',
            'source',
            'ml_class',
            'state',
            'confidence',
        )


class ClassificationCreateSerializer(serializers.ModelSerializer):
    """
    Specific serializer for the manual classification creation
    """
    class Meta:
        model = Classification
        fields = (
            'element',
            'ml_class',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['element'].queryset = Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user))


class TranscriptionSerializer(serializers.ModelSerializer):
    """
    Serialises a Transcription
    issued from a search
    """

    type = EnumField(TranscriptionType)
    zone = ZoneSerializer()
    source = DataSourceSerializer()

    class Meta:
        model = Transcription
        fields = (
            'id',
            'type',
            'text',
            'score',
            'zone',
            'source',
        )


class TranscriptionCreateSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    """
    element = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    source = serializers.PrimaryKeyRelatedField(
        queryset=DataSource.objects.filter(type=MLToolType.Recognizer),
    )
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['element'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class TranscriptionBulkSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (used by serializer below)
    Note: no element !
    """
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)


class TranscriptionsSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (uses serializer above) on a common parent
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    recognizer = MLToolField(MLToolType.Recognizer)
    transcriptions = TranscriptionBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class ClassificationBulkSerializer(serializers.Serializer):
    """
    Single classification serializer for bulk insertion
    Cannot use ModelSerializer as they become read-only when nested
    """
    class_name = serializers.CharField(source='ml_class')
    confidence = serializers.FloatField(min_value=0, max_value=1)


class ClassificationsSerializer(serializers.Serializer):
    """
    Insert N classifications on a single element from a single ML tool
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.all(),
    )
    classifier = MLToolField(MLToolType.Classifier)
    classifications = ClassificationBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )

    def validate(self, data):
        data = super().validate(data)
        corpus_id = data['parent'].corpus.id
        for classification in data['classifications']:
            # Retrieve or create required ml_class
            classification['ml_class'], _ = MLClass.objects.get_or_create(
                name=classification['ml_class'],
                corpus_id=corpus_id
            )
        return data
