from django.urls import reverse
from rest_framework import status
from unittest.mock import patch
from arkindex.project.tests import FixtureAPITestCase
from arkindex.documents.models import Corpus, DataSource, MLClass


class TestBulkClassification(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.src = DataSource.objects.get(slug='test')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.ml_get_patch = patch('arkindex.project.serializer_fields.MLTool.get')
        cls.ml_get_mock = cls.ml_get_patch.start()
        cls.ml_get_mock.return_value.type = cls.src.type
        cls.ml_get_mock.return_value.name = cls.src.name
        cls.ml_get_mock.return_value.slug = cls.src.slug
        cls.ml_get_mock.return_value.version = cls.src.revision
        cls.src.internal = True
        cls.src.save()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.ml_get_patch.stop()
        cls.src.internal = False
        cls.src.save()

    def create_classifications_data(self, classifications, parent=None, slug=None):
        return {
            "parent": parent or str(self.page.id),
            "classifier": slug or self.src.slug,
            "classifications": classifications
        }

    def test_bulk_classification_requires_login(self):
        response = self.client.post(reverse('api:classification-bulk'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_classification_wrong_acl(self):
        self.client.force_login(self.user)
        private_page = self.private_corpus.elements.create(
            type=self.private_corpus.types.create(slug='page'),
        )
        response = self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data(
                [{"class_name": 'dog', "confidence": 0.99}],
                parent=str(private_page.id)
            )
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'parent': ['Invalid pk "{}" - object does not exist.'.format(private_page.id)]
            }
        )

    def test_bulk_classification(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data([
                {
                    "class_name": 'dog',
                    "confidence": 0.99,
                },
                {
                    "class_name": 'cat',
                    "confidence": 0.42,
                }
            ])
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence')),
            [
                ('dog', 0.99),
                ('cat', 0.42),
            ],
        )

    def test_bulk_classification_create_ml_class(self):
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 0)
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data(
                [{"class_name": 'dog', "confidence": 0.99}],
            )
        )
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 1)

    def test_bulk_classification_get_existing_ml_class(self):
        MLClass.objects.create(name="cat", corpus=self.corpus)
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data(
                [{"class_name": 'cat', "confidence": 0.42}],
            )
        )
        self.assertEqual(MLClass.objects.filter(name="cat", corpus=self.corpus).count(), 1)

    def test_bulk_classification_delete(self):
        """
        Test the bulk classification API deletes previous classifications
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data([
                {
                    "class_name": 'dog',
                    "confidence": 0.99,
                },
                {
                    "class_name": 'cat',
                    "confidence": 0.42,
                }
            ])
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(
            reverse('api:classification-bulk'),
            format='json',
            data=self.create_classifications_data([
                {
                    "class_name": 'doggo',
                    "confidence": 0.5,
                },
                {
                    "class_name": 'catte',
                    "confidence": 0.85,
                }
            ])
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence')),
            [
                ('doggo', 0.5),
                ('catte', 0.85),
            ],
        )
