from django.urls import reverse
from rest_framework import status
from arkindex.documents.models import Classification, DataSource, MLClass
from arkindex.project.tests import FixtureAPITestCase


class TestClasses(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name='text', corpus=cls.corpus)
        cls.cover = MLClass.objects.create(name='cover', corpus=cls.corpus)

    def test_list(self):
        """
        Test listing results alpha-ordered
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover"
                },
                {
                    "id": str(self.text.pk),
                    "name": "text"
                },
            ]
        })

        # Add other classes to ensure ordering is preserved
        self.white_page = MLClass.objects.create(name='white-page', corpus=self.corpus)
        self.image = MLClass.objects.create(name='image', corpus=self.corpus)
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 4,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover"
                },
                {
                    "id": str(self.image.pk),
                    "name": "image"
                },
                {
                    "id": str(self.text.pk),
                    "name": "text"
                },
                {
                    "id": str(self.white_page.pk),
                    "name": "white-page"
                },
            ]
        })

    def test_list_with_folder_filter(self):
        """
        Test listing results alpha-ordered with a folder filter
        """
        self.client.force_login(self.user)
        folder = self.corpus.elements.create(
            type=self.corpus.types.get(slug='volume'),
            name="Custom folder",
        )
        data_source = DataSource.objects.get(slug='test')

        # Test with a newly created folder: must not return any MLClass
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {'folder': folder.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 0,
            "number": 1,
            "next": None,
            "previous": None,
            "results": []
        })

        # Attach page to the folder: must not return any MLClass without classification
        page = self.corpus.elements.create(
            type=self.corpus.types.get(slug='page'),
            name='Custom page',
        )
        page.add_parent(folder)
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {'folder': folder.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 0,
            "number": 1,
            "next": None,
            "previous": None,
            "results": []
        })

        # Add classification: must return class
        Classification.objects.create(element=page, source=data_source, ml_class=self.text)
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {'folder': folder.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.text.pk),
                    "name": "text"
                }
            ]
        })

        # Add other classifications to ensure ordering is preserved
        self.white_page = MLClass.objects.create(name='white-page', corpus=self.corpus)
        self.image = MLClass.objects.create(name='image', corpus=self.corpus)
        Classification.objects.create(element=page, source=data_source, ml_class=self.white_page)
        Classification.objects.create(element=page, source=data_source, ml_class=self.image)
        response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {'folder': folder.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 3,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.image.pk),
                    "name": "image"
                },
                {
                    "id": str(self.text.pk),
                    "name": "text"
                },
                {
                    "id": str(self.white_page.pk),
                    "name": "white-page"
                }
            ]
        })

    def test_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_update(self):
        self.client.force_login(self.user)
        response = self.client.put(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        response = self.client.patch(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
