from django.urls import reverse
from rest_framework import status
from arkindex.documents.models import Corpus, ElementPath
from arkindex.project.tests import FixtureAPITestCase


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='the_other_one')
        # Create types
        cls.house_type = cls.corpus.types.create(slug='house', display_name='House', folder=True)
        cls.room_type = cls.corpus.types.create(slug='room', display_name='Room', folder=True)
        cls.furniture_type = cls.corpus.types.create(slug='furniture', display_name='Furniture')
        cls.leg_type = cls.corpus.types.create(slug='color', display_name='Furniture Leg', hidden=True)

    def setUp(self):
        super().setUp()
        # Populate elements
        self.house = self.corpus.elements.create(type_id=self.house_type.id, name='House')
        self.room = self.corpus.elements.create(type_id=self.room_type.id, name='Kitchen')
        self.desk = self.corpus.elements.create(type_id=self.furniture_type.id, name='Desk')
        self.default_kwargs = {
            'child': str(self.desk.id),
            'parent': str(self.room.id)
        }

    def link_elements(self):
        self.desk.add_parent(self.room)
        self.room.add_parent(self.house)

    def test_element_path_create(self):
        self.assertEqual(ElementPath.objects.filter(element_id=self.desk.id).count(), 0)
        self.client.force_login(self.user)
        # Link desk to its parent room
        response = self.client.post(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(ElementPath.objects.filter(element_id=self.desk.id).count(), 1)
        self.assertListEqual(
            ElementPath.objects.get(element_id=self.desk.id).path,
            [self.room.id]
        )
        # Link room to its parent house
        self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.room.id),
            'parent': str(self.house.id)
        }))
        self.assertListEqual(
            ElementPath.objects.get(element_id=self.desk.id).path,
            [self.house.id, self.room.id]
        )

    def test_create_forbidden(self):
        response = self.client.post(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_acl(self):
        self.client.force_login(self.user)
        self.desk.corpus = self.private_corpus
        self.desk.save()
        response = self.client.post(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'child': ['Invalid pk "{}" - object does not exist.'.format(self.desk.id)]}
        )

    def test_different_corpus(self):
        other_corpus = Corpus.objects.create(name='private')
        other_corpus.corpus_right.create(user=self.user, can_write=True)
        theater = other_corpus.elements.create(type_id=self.room_type.id, name='Theater')

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.desk.id),
            'parent': str(theater.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': ["Parent is not from corpus '{}'".format(self.room.corpus.name)]}
        )

    def test_own_parent(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.room.id),
            'parent': str(self.room.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': ['A child cannot be its own parent']}
        )

    def test_parent_already_exists(self):
        self.link_elements()
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': ["'{}' is already a parent of '{}'".format(self.room.id, self.desk.id)]}
        )

    def test_cycles(self):
        self.link_elements()
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.house.id),
            'parent': str(self.room.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': ["'{}' is a child of element '{}'".format(self.room.id, self.house.id)]}
        )

    def test_folder_child_folder_parent(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.room.id),
            'parent': str(self.house.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_hidden_child_folder_parent(self):
        self.client.force_login(self.user)
        desk_leg = self.corpus.elements.create(type_id=self.leg_type.id, name='desk leg')
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(desk_leg.id),
            'parent': str(self.room.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': ["Element '{}' has folder property thus cannot contain hidden elements".format(self.room.id)]}
        )

    def test_folder_child_normal_parent(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(self.room.id),
            'parent': str(self.desk.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': [
                "Element '{}' does not have folder property and may "
                "contain hidden elements only".format(self.desk.id)
            ]}
        )

    def test_hidden_child_normal_parent(self):
        self.client.force_login(self.user)
        desk_leg = self.corpus.elements.create(type_id=self.leg_type.id, name='desk leg')
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(desk_leg.id),
            'parent': str(self.desk.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_normal_child_normal_parent(self):
        self.client.force_login(self.user)
        chair = self.corpus.elements.create(type_id=self.furniture_type.id, name='Chair')
        response = self.client.post(reverse('api:element-parent', kwargs={
            'child': str(chair.id),
            'parent': str(self.desk.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'parent': [
                "Element '{}' does not have folder property and may "
                "contain hidden elements only".format(self.desk.id)
            ]}
        )

    def test_delete_forbidden(self):
        response = self.client.delete(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_acl(self):
        self.client.force_login(self.user)
        self.desk.corpus = self.private_corpus
        self.desk.save()
        response = self.client.delete(reverse('api:element-parent', kwargs=self.default_kwargs))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'child': ['Invalid pk "{}" - object does not exist.'.format(self.desk.id)]}
        )

    def test_delete_relation(self):
        self.link_elements()
        self.client.force_login(self.user)
        self.assertListEqual(
            ElementPath.objects.get(element_id=self.desk.id).path,
            [self.house.id, self.room.id]
        )
        response = self.client.delete(reverse('api:element-parent', kwargs={
            'child': str(self.room.id),
            'parent': str(self.house.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertListEqual(
            ElementPath.objects.get(element_id=self.desk.id).path,
            [self.room.id]
        )

    def test_delete_unexisting_relation(self):
        self.client.force_login(self.user)
        self.assertEqual(ElementPath.objects.filter(element_id=self.desk.id).count(), 0)
        response = self.client.delete(reverse('api:element-parent', kwargs={
            'child': str(self.desk.id),
            'parent': str(self.room.id)
        }))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
