from pathlib import Path
from django.urls import reverse
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import TranscriptionType, EntityType
from arkindex.documents.models import Entity, EntityRole, EntityLink, Transcription, TranscriptionEntity
from arkindex.documents.pagexml import PageXmlParser

FIXTURES = Path(__file__).absolute().parent / 'pagexml_samples'
TRANSKRIBUS_ROLE = Path(__file__).absolute().parent / '../fixtures/roles_transkribus.yml'


class TestPageXml(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 2, page 1r')
        cls.types = {
            '_nom': EntityType.Person.value,
            '_prenom': EntityType.Person.value,
            '_lieu_residence': EntityType.Location.value,
            '_profession': EntityType.Subject.value,
        }

    def all_tags(self, block):
        for line in block['lines']:
            for tag in line['tags']:
                yield tag

    def test_pagexml_import_requires_login(self):
        with (FIXTURES / 'transcript.xml').open() as f:
            resp = self.client.post(
                reverse('api:pagexml-transcriptions', kwargs={'pk': str(self.page.id)}),
                data=f.read(),
                content_type='application/xml',
            )
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_pagexml_import(self):
        self.assertFalse(self.page.transcriptions.exists())
        self.client.force_login(self.user)
        with (FIXTURES / 'transcript.xml').open() as f:
            resp = self.client.post(
                reverse('api:pagexml-transcriptions', kwargs={'pk': str(self.page.id)}),
                data=f.read(),
                content_type='application/xml',
            )
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(self.page.transcriptions.values_list('type', 'text'), [
            (TranscriptionType.Paragraph, 'B .1\nLouis Joseph\nPierre Siméon\nLemieux'),
            (TranscriptionType.Paragraph, 'Le onze janvier mil neuf centsept\n'
                                          'nous prêtre soussigné avons baptisé Louis'),
            (TranscriptionType.Line, 'B .1'),
            (TranscriptionType.Line, 'Louis Joseph'),
            (TranscriptionType.Line, 'Pierre Siméon'),
            (TranscriptionType.Line, 'Lemieux'),
            (TranscriptionType.Line, 'Le onze janvier mil neuf centsept'),
            (TranscriptionType.Line, 'nous prêtre soussigné avons baptisé Louis'),
        ])
        # All transcriptions have a score of 100%
        self.assertFalse(self.page.transcriptions.exclude(score=1.0).exists())

    def test_pagexml_import_requires_zone(self):
        volume = self.corpus.elements.get(name='Volume 1')
        self.assertIsNone(volume.zone)

        self.client.force_login(self.user)
        with (FIXTURES / 'transcript.xml').open() as f:
            resp = self.client.post(
                reverse('api:pagexml-transcriptions', kwargs={'pk': str(volume.id)}),
                data=f.read(),
                content_type='application/xml',
            )
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

    def test_pagexml_import_any_element(self):
        surface = self.corpus.elements.get(name='Surface A')
        self.client.force_login(self.user)
        with (FIXTURES / 'transcript.xml').open() as f:
            resp = self.client.post(
                reverse('api:pagexml-transcriptions', kwargs={'pk': str(surface.id)}),
                data=f.read(),
                content_type='application/xml',
            )
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(surface.transcriptions.values_list('type', 'text'), [
            (TranscriptionType.Paragraph, 'B .1\nLouis Joseph\nPierre Siméon\nLemieux'),
            (TranscriptionType.Paragraph, 'Le onze janvier mil neuf centsept\n'
                                          'nous prêtre soussigné avons baptisé Louis'),
            (TranscriptionType.Line, 'B .1'),
            (TranscriptionType.Line, 'Louis Joseph'),
            (TranscriptionType.Line, 'Pierre Siméon'),
            (TranscriptionType.Line, 'Lemieux'),
            (TranscriptionType.Line, 'Le onze janvier mil neuf centsept'),
            (TranscriptionType.Line, 'nous prêtre soussigné avons baptisé Louis'),
        ])
        # All transcriptions have a score of 100%
        self.assertFalse(surface.transcriptions.exclude(score=1.0).exists())

    def test_pagexml_create_blocks(self):
        parser = PageXmlParser(FIXTURES / 'create_blocks.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)

        text_region_0 = 'Le onze janvier mil neuf centsept\nnous prêtre soussigné avons baptisé Louis'
        text_region_1 = 'B .1\nLouis Joseph\nPierre Siméon\nLemieux'
        transcription_id_0 = Transcription.objects.get(text=text_region_0).id
        transcription_id_1 = Transcription.objects.get(text=text_region_1).id

        self.assertEqual(
            blocks,
            [
                {
                    'reading_order': 0,
                    'lines': [
                        {
                            'reading_order': 0,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '0',
                                    'reading_order': 0,
                                    'transcription_id': transcription_id_0,
                                    'offset_in_region': 0
                                },
                                {
                                    'name': '_date',
                                    'offset': '3',
                                    'length': '30',
                                    '_enregistrement': '1',
                                    'value': 'onze janvier mil neuf centsept',
                                    'reading_order': 0,
                                    'transcription_id': transcription_id_0,
                                    'offset_in_region': 3
                                }
                            ],
                            'length': 33
                        },
                        {
                            'reading_order': 1,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '1',
                                    'reading_order': 1,
                                    'transcription_id': transcription_id_0,
                                    'offset_in_region': 34
                                },
                                {
                                    'name': '_prenom',
                                    'offset': '36',
                                    'length': '5',
                                    'continued': 'true',
                                    '_role': 'sujet',
                                    'value': 'Louis',
                                    'reading_order': 1,
                                    'transcription_id': transcription_id_0,
                                    'offset_in_region': 70
                                }
                            ], 'length': 41
                        }
                    ]
                },
                {
                    'reading_order': 1,
                    'lines': [
                        {
                            'reading_order': 0,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '0',
                                    'reading_order': 0,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 0
                                },
                                {
                                    'name': 'structure',
                                    'type': 'ref',
                                    'reading_order': 0,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 0
                                }
                            ],
                            'length': 4
                        },
                        {
                            'reading_order': 1,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '1',
                                    'reading_order': 1,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 5
                                },
                                {
                                    'name': '_prenom',
                                    'offset': '0',
                                    'length': '12',
                                    'continued': 'true',
                                    '_role': 'sujet',
                                    'value': 'Louis Joseph',
                                    'reading_order': 1,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 5
                                }
                            ],
                            'length': 12
                        },
                        {
                            'reading_order': 2,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '2',
                                    'reading_order': 2,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 18
                                },
                                {
                                    'name': '_prenom',
                                    'offset': '0',
                                    'length': '13',
                                    'continued': 'true',
                                    '_role': 'sujet',
                                    'value': 'Pierre Siméon',
                                    'reading_order': 2,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 18
                                }
                            ],
                            'length': 13
                        }, {
                            'reading_order': 3,
                            'tags': [
                                {
                                    'name': 'readingOrder',
                                    'index': '3',
                                    'reading_order': 3,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 32
                                },
                                {
                                    'name': '_nom',
                                    'offset': '0',
                                    'length': '7',
                                    'value': 'Lemieux',
                                    'reading_order': 3,
                                    'transcription_id': transcription_id_1,
                                    'offset_in_region': 32
                                }
                            ],
                            'length': 7
                        }
                    ]
                }
            ]
        )

    def test_pagexml_regroup(self):
        parser = PageXmlParser(FIXTURES / 'regroup.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(len(blocks[0]['lines']), 5)
        self.assertEqual(len(blocks[1]['lines']), 5)
        tags = list(filter(lambda tag: tag['name'][0] == '_', self.all_tags(blocks[0])))

        tags = parser.regroup(tags)
        tags = parser.cesure(tags)

        text_region = 'Nous avons bap-\ntisé Marie-Evilina, fille de\nCamille Thibeault\net de Ermi-\nnise Champeau.\n'
        transcription_id = Transcription.objects.get(text=text_region).id

        self.assertEqual(
            tags,
            [
                {
                    'name': '_prenom',
                    'offset': '5',
                    'length': 13,
                    '_role': 'sujet',
                    'value': 'Marie-Evilina',
                    'reading_order': 1,
                    'transcription_id': transcription_id,
                    'offset_in_region': 21,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '5',
                            'length': '13',
                            '_role': 'sujet',
                            'value': 'Marie-Evilina',
                            'reading_order': 1,
                            'transcription_id': transcription_id,
                            'offset_in_region': 21
                        }
                    ]
                },
                {
                    'name': '_prenom',
                    'offset': '0',
                    'length': 7,
                    '_role': 'père',
                    'value': 'Camille',
                    'reading_order': 2,
                    'transcription_id': transcription_id,
                    'offset_in_region': 45,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '7',
                            '_role': 'père',
                            'value': 'Camille',
                            'reading_order': 2,
                            'transcription_id': transcription_id,
                            'offset_in_region': 45
                        }
                    ]
                },
                {
                    'name': '_nom',
                    'offset': '8',
                    'length': 9,
                    'value': 'Thibeault',
                    'reading_order': 2,
                    'transcription_id': transcription_id,
                    'offset_in_region': 53,
                    'items': [
                        {
                            'name': '_nom',
                            'offset': '8',
                            'length': '9',
                            'value': 'Thibeault',
                            'reading_order': 2,
                            'transcription_id': transcription_id,
                            'offset_in_region': 53
                        }
                    ]
                },
                {
                    'name': '_prenom',
                    'offset': '6',
                    'length': 10,
                    'continued': 'true',
                    '_role': 'mère',
                    'value': 'Erminise',
                    'reading_order': 4,
                    'transcription_id': transcription_id,
                    'offset_in_region': 69,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '6',
                            'length': '5',
                            'continued': 'true',
                            '_role': 'mère',
                            'value': 'Ermi-',
                            'reading_order': 3,
                            'transcription_id': transcription_id,
                            'offset_in_region': 69
                        },
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '4',
                            'continued': 'true',
                            '_role': 'mère',
                            'value': 'nise',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 75
                        }
                    ]
                },
                {
                    'name': '_nom',
                    'offset': '5',
                    'length': 8,
                    'value': 'Champeau',
                    'reading_order': 4,
                    'transcription_id': transcription_id,
                    'offset_in_region': 80,
                    'items': [
                        {
                            'name': '_nom',
                            'offset': '5',
                            'length': '8',
                            'value': 'Champeau',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 80
                        }
                    ]
                }
            ]
        )

        tags = list(filter(lambda tag: tag['name'][0] == '_', self.all_tags(blocks[1])))
        tags = parser.regroup(tags)
        tags = parser.cesure(tags)

        text_region = 'S. 10\nDragon\nMarie\nEmillia\nElmire'
        transcription_id = Transcription.objects.get(text=text_region).id

        self.assertEqual(
            tags,
            [
                {
                    'name': '_nom',
                    'offset': '0',
                    'length': 6,
                    'value': 'Dragon',
                    'reading_order': 1,
                    'transcription_id': transcription_id,
                    'offset_in_region': 6,
                    'items': [
                        {
                            'name': '_nom',
                            'offset': '0',
                            'length': '6',
                            'value': 'Dragon',
                            'reading_order': 1,
                            'transcription_id': transcription_id,
                            'offset_in_region': 6
                        }
                    ]
                },
                {
                    'name': '_prenom',
                    'offset': '0',
                    'length': 20,
                    'continued': 'true',
                    '_role': 'sujet',
                    'value': 'Marie Emillia Elmire',
                    'reading_order': 4,
                    'transcription_id': transcription_id,
                    'offset_in_region': 13,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '5',
                            'continued': 'true',
                            '_role': 'sujet',
                            'value': 'Marie',
                            'reading_order': 2,
                            'transcription_id': transcription_id,
                            'offset_in_region': 13
                        },
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '7',
                            'continued': 'true',
                            '_role': 'sujet',
                            'value': 'Emillia',
                            'reading_order': 3,
                            'transcription_id': transcription_id,
                            'offset_in_region': 19
                        },
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '6',
                            'continued': 'true',
                            '_role': 'sujet',
                            'value': 'Elmire',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 27
                        }
                    ]
                }
            ]
        )

    def test_pagexml_first_and_last_name(self):
        parser = PageXmlParser(FIXTURES / 'first_and_last_name.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        self.assertEqual(len(blocks), 3)
        self.assertEqual(len(blocks[0]['lines']), 7)
        self.assertEqual(len(blocks[1]['lines']), 6)
        self.assertEqual(len(blocks[2]['lines']), 6)
        tags = list(filter(lambda tag: tag['name'][0] == '_', self.all_tags(blocks[0])))

        tags = parser.regroup(tags)
        self.assertEqual(len(tags), 3)
        tags = parser.first_and_last_name(tags)
        tags = parser.cesure(tags)

        text_region = 'B. 31\nChaloux\n( ) Gérard\nHenri\nP.107\nv N.H.\nT.O.N. PtreCuré'
        transcription_id = Transcription.objects.get(text=text_region).id

        self.assertEqual(
            tags,
            [
                {
                    'name': '_prenom',
                    'offset': '4',
                    'length': 20,
                    'continued': 'true',
                    '_role': 'sujet',
                    'value': 'Gérard Henri Chaloux',
                    'reading_order': 3,
                    'transcription_id': transcription_id,
                    'offset_in_region': 18,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '4',
                            'length': '6',
                            'continued': 'true',
                            '_role': 'sujet',
                            'value': 'Gérard',
                            'reading_order': 2,
                            'transcription_id': transcription_id,
                            'offset_in_region': 18
                        },
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '5',
                            'continued': 'true',
                            '_role': 'sujet',
                            'value': 'Henri',
                            'reading_order': 3,
                            'transcription_id': transcription_id,
                            'offset_in_region': 25
                        },
                        {
                            'name': '_nom',
                            'offset': '0',
                            'length': '7',
                            'value': 'Chaloux',
                            'reading_order': 1,
                            'transcription_id': transcription_id,
                            'offset_in_region': 6
                        }
                    ],
                    '_role': 'sujet'
                },
                {
                    'name': '_lieu_residence',
                    'offset': '2',
                    'length': 4,
                    'value': 'N.H.',
                    'reading_order': 5,
                    'transcription_id': transcription_id,
                    'offset_in_region': 39,
                    'items': [
                        {
                            'name': '_lieu_residence',
                            'offset': '2',
                            'length': '4',
                            'value': 'N.H.',
                            'reading_order': 5,
                            'transcription_id': transcription_id,
                            'offset_in_region': 39
                        }
                    ]
                }
            ]
        )

        tags = list(filter(lambda tag: tag['name'][0] == '_', self.all_tags(blocks[1])))
        tags = parser.regroup(tags)
        self.assertEqual(len(tags), 5)
        tags = parser.first_and_last_name(tags)
        tags = parser.cesure(tags)

        text_region = 'Nous avons\ninhumé\nle corps d\'un garçon issu \nde Joseph Du-\nmoulin, et \nde Valeda Paquette.'
        transcription_id = Transcription.objects.get(text=text_region).id

        self.assertEqual(
            tags,
            [
                {
                    'name': '_prenom',
                    'offset': '14',
                    'length': 17,
                    '_role': 'sujet',
                    'value': 'garçon Dumoulin',
                    'reading_order': 3,
                    'transcription_id': transcription_id,
                    'offset_in_region': 32,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '14',
                            'length': '6',
                            '_role': 'sujet',
                            'value': 'garçon',
                            'reading_order': 3,
                            'transcription_id': transcription_id,
                            'offset_in_region': 32
                        },
                        {
                            'name': '_nom',
                            'offset': '10',
                            'length': '3',
                            'continued': 'true',
                            'value': 'Du-',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 54
                        },
                        {
                            'name': '_nom',
                            'offset': '0',
                            'length': '6',
                            'continued': 'true',
                            'value': 'moulin',
                            'reading_order': 5,
                            'transcription_id': transcription_id,
                            'offset_in_region': 58
                        }
                    ]
                },
                {
                    'name': '_prenom',
                    'offset': '3',
                    'length': 17,
                    '_role': 'père',
                    'value': 'Joseph Dumoulin',
                    'reading_order': 4,
                    'transcription_id': transcription_id,
                    'offset_in_region': 47,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '3',
                            'length': '6',
                            '_role': 'père',
                            'value': 'Joseph',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 47
                        },
                        {
                            'name': '_nom',
                            'offset': '10',
                            'length': '3',
                            'continued': 'true',
                            'value': 'Du-',
                            'reading_order': 4,
                            'transcription_id': transcription_id,
                            'offset_in_region': 54
                        },
                        {
                            'name': '_nom',
                            'offset': '0',
                            'length': '6',
                            'continued': 'true',
                            'value': 'moulin',
                            'reading_order': 5,
                            'transcription_id': transcription_id,
                            'offset_in_region': 58
                        }
                    ]
                },
                {
                    'name': '_prenom',
                    'offset': '3',
                    'length': 15,
                    '_role': 'mère',
                    'value': 'Valeda Paquette',
                    'reading_order': 6,
                    'transcription_id': transcription_id,
                    'offset_in_region': 73,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '3',
                            'length': '6',
                            '_role': 'mère',
                            'value': 'Valeda',
                            'reading_order': 6,
                            'transcription_id': transcription_id,
                            'offset_in_region': 73
                        },
                        {
                            'name': '_nom',
                            'offset': '10',
                            'length': '8',
                            'value': 'Paquette',
                            'reading_order': 6,
                            'transcription_id': transcription_id,
                            'offset_in_region': 80
                        }
                    ]
                },
            ]
        )

        tags = list(filter(lambda tag: tag['name'][0] == '_', self.all_tags(blocks[2])))
        tags = parser.regroup(tags)
        self.assertEqual(len(tags), 2)
        tags = parser.first_and_last_name(tags)
        tags = parser.cesure(tags)

        text_region = 'S. 12\nLapalme\nCharles-Albert\nP.107\nS. Chycoine\nvoir folio 10'
        transcription_id = Transcription.objects.get(text=text_region).id

        self.assertEqual(
            tags,
            [
                {
                    'name': '_prenom',
                    'offset': '0',
                    'length': 22,
                    '_role': 'sujet',
                    'value': 'Charles-Albert Lapalme',
                    'reading_order': 2,
                    'transcription_id': transcription_id,
                    'offset_in_region': 14,
                    'items': [
                        {
                            'name': '_prenom',
                            'offset': '0',
                            'length': '14',
                            '_role': 'sujet',
                            'value': 'Charles-Albert',
                            'reading_order': 2,
                            'transcription_id': transcription_id,
                            'offset_in_region': 14
                        },
                        {
                            'name': '_nom',
                            'offset': '0',
                            'length': '7',
                            'value': 'Lapalme',
                            'reading_order': 1,
                            'transcription_id': transcription_id,
                            'offset_in_region': 6
                        }
                    ]
                }
            ]
        )

    def test_merge(self):
        parser = PageXmlParser(FIXTURES / 'merge.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        text_region_0 = 'Nous\navons baptisé\nMarie-Alice-Rosilda, née\nde Antoine Landry\net de Zelpha\nBeauchesne.'
        text_region_1 = 'B. 32\nLandry\nMarie-Alice\nRosilda\nP.108'
        transcription_id_0 = Transcription.objects.get(text=text_region_0).id
        transcription_id_1 = Transcription.objects.get(text=text_region_1).id

        self.assertEqual(
            blocks,
            {
                0: [
                    {
                        'name': '_prenom',
                        'offset': '0',
                        'length': 26,
                        '_role': 'sujet',
                        'value': 'Marie-Alice-Rosilda Landry',
                        'reading_order': 2,
                        'transcription_id': transcription_id_0,
                        'offset_in_region': 42,
                        'items': [
                            {
                                'name': '_prenom',
                                'offset': '0',
                                'length': '19',
                                '_role': 'sujet',
                                'value': 'Marie-Alice-Rosilda',
                                'reading_order': 2,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 42
                            },
                            {
                                'name': '_nom',
                                'offset': '11',
                                'length': '6',
                                'value': 'Landry',
                                'reading_order': 3,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 78
                            }
                        ]
                    },
                    {
                        'name': '_prenom',
                        'offset': '3',
                        'length': 14,
                        '_role': 'père',
                        'value': 'Antoine Landry',
                        'reading_order': 3,
                        'transcription_id': transcription_id_0,
                        'offset_in_region': 70,
                        'items': [
                            {
                                'name': '_prenom',
                                'offset': '3',
                                'length': '7',
                                '_role': 'père',
                                'value': 'Antoine',
                                'reading_order': 3,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 70
                            },
                            {
                                'name': '_nom',
                                'offset': '11',
                                'length': '6',
                                'value': 'Landry',
                                'reading_order': 3,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 78
                            }
                        ]
                    },
                    {
                        'name': '_prenom',
                        'offset': '6',
                        'length': 17,
                        '_role': 'mère',
                        'value': 'Zelpha Beauchesne',
                        'reading_order': 4,
                        'transcription_id': transcription_id_0,
                        'offset_in_region': 91,
                        'items': [
                            {
                                'name': '_prenom',
                                'offset': '6',
                                'length': '6',
                                '_role': 'mère',
                                'value': 'Zelpha',
                                'reading_order': 4,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 91
                            },
                            {
                                'name': '_nom',
                                'offset': '0',
                                'length': '10',
                                'value': 'Beauchesne',
                                'reading_order': 5,
                                'transcription_id': transcription_id_0,
                                'offset_in_region': 98
                            }
                        ]
                    }
                ],
                1: [
                    {
                        'name': '_prenom',
                        'offset': '0',
                        'length': 26,
                        'continued': 'true',
                        '_role': 'sujet',
                        'value': 'Marie-Alice Rosilda Landry',
                        'reading_order': 3,
                        'transcription_id': transcription_id_1,
                        'offset_in_region': 13,
                        'items': [
                            {
                                'name': '_prenom',
                                'offset': '0',
                                'length': '11',
                                'continued': 'true',
                                '_role': 'sujet',
                                'value': 'Marie-Alice',
                                'reading_order': 2,
                                'transcription_id': transcription_id_1,
                                'offset_in_region': 13
                            },
                            {
                                'name': '_prenom',
                                'offset': '0',
                                'length': '7',
                                'continued': 'true',
                                '_role': 'sujet',
                                'value': 'Rosilda',
                                'reading_order': 3,
                                'transcription_id': transcription_id_1,
                                'offset_in_region': 25
                            },
                            {
                                'name': '_nom',
                                'offset': '0',
                                'length': '6',
                                'value': 'Landry',
                                'reading_order': 1,
                                'transcription_id': transcription_id_1,
                                'offset_in_region': 6
                            }
                        ]
                    }
                ]
            }
        )

    def test_create_entities(self):
        parser = PageXmlParser(FIXTURES / 'create_entities.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        entities = []
        for block in blocks:
            entities += parser.create_entities(blocks[block], self.page, self.types, 0.90)[0]

        nb_entities = Entity.objects.filter(corpus_id=self.page.corpus.id).count()
        self.assertEqual(len(entities), nb_entities)

    def test_create_roles(self):
        parser = PageXmlParser(FIXTURES / 'merge.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        for block in blocks.values():
            entities, tags = parser.create_entities(block, self.page, self.types, 0.90)
            parser.create_roles(tags, self.page, TRANSKRIBUS_ROLE, 0.90)

        nb_roles = EntityRole.objects.filter(corpus_id=self.page.corpus.id).count()
        self.assertEqual(2, nb_roles)

    def test_create_links(self):
        parser = PageXmlParser(FIXTURES / 'merge.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        for block in blocks.values():
            entities, tags = parser.create_entities(block, self.page, self.types, 0.90)
            tags = parser.create_roles(tags, self.page, TRANSKRIBUS_ROLE, 0.90)
            parser.create_links(tags, TRANSKRIBUS_ROLE, 0.90)

        nb_entities = Entity.objects.filter(corpus_id=self.page.corpus.id).count()
        nb_roles = EntityRole.objects.filter(corpus_id=self.page.corpus.id).count()
        nb_links = EntityLink.objects.filter(role__corpus__id=self.page.corpus.id).count()
        self.assertEqual(3, nb_entities)
        self.assertEqual(2, nb_roles)
        self.assertEqual(2, nb_links)

    def test_create_objects(self):
        parser = PageXmlParser(FIXTURES / 'merge.xml')
        transcriptions = parser.save(self.page)
        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)
        entities_id = parser.create_objects(blocks, self.page, TRANSKRIBUS_ROLE, self.types)

        nb_entities = Entity.objects.filter(corpus_id=self.page.corpus.id).count()
        nb_roles = EntityRole.objects.filter(corpus_id=self.page.corpus.id).count()
        nb_links = EntityLink.objects.filter(role__corpus__id=self.page.corpus.id).count()
        nb_transcription_entity = TranscriptionEntity.objects.filter(entity__id__in=entities_id).count()
        self.assertEqual(3, nb_entities)
        self.assertEqual(2, nb_roles)
        self.assertEqual(2, nb_links)
        self.assertEqual(5, nb_transcription_entity)

    def test_cesure(self):
        parser = PageXmlParser(FIXTURES / 'merge.xml')

        personne = [
            {
                'items': [{'value': ' Marie    '}, {'value': ' Anne '}]
            }
        ]
        completed_personne = parser.cesure(personne)
        self.assertEqual(completed_personne[0]['value'], 'Marie Anne')

        personne = [
            {
                'items': [{'value': ' Mari    '}, {'value': ' anne '}]
            }
        ]
        completed_personne = parser.cesure(personne)
        self.assertEqual(completed_personne[0]['value'], 'Marianne')

        personne = [
            {
                'items': [{'value': '    Marie-  - '}, {'value': '---Anne '}]
            }
        ]
        completed_personne = parser.cesure(personne)
        self.assertEqual(completed_personne[0]['value'], 'Marie-Anne')

        personne = [
            {
                'items': [{'value': '  Mari --'}, {'value': '- - anne   '}]
            }
        ]
        completed_personne = parser.cesure(personne)
        self.assertEqual(completed_personne[0]['value'], 'Marianne')
