from unittest.mock import patch
from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex.documents.models import DataSource, Region


class TestRegionsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.ds = DataSource.objects.create(
            type=MLToolType.DLAnalyser,
            slug='tasty',
            name='Tasty analyser',
            revision='2',
            internal=True,
        )
        cls.ml_class = cls.corpus.ml_classes.create(name='bacon')

        cls.z1 = cls.page.zone.image.zones.create(polygon=Polygon.from_coords(100, 100, 50, 50))
        cls.z2 = cls.page.zone.image.zones.create(polygon=Polygon.from_coords(200, 200, 50, 50))

        cls.r1 = cls.ds.regions.create(ml_class=cls.ml_class, confidence=0.8, zone=cls.z1)
        cls.r2 = cls.ds.regions.create(ml_class=cls.ml_class, confidence=0.5, zone=cls.z2)

        cls.r1.region_elements.create(element=cls.page, ordering=0)
        cls.r2.region_elements.create(element=cls.page, ordering=1)

    def test_region_details(self):
        response = self.client.get(reverse('api:region-details', kwargs={'pk': str(self.r1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            "id": str(self.r1.id),
            "source": {
                "id": str(self.ds.id),
                "type": "dla",
                "slug": "tasty",
                "name": "Tasty analyser",
                "revision": "2",
                "internal": True,
            },
            "ml_class": {
                "id": str(self.ml_class.id),
                "name": "bacon",
            },
            "confidence": 0.8,
            "zone": {
                "id": str(self.z1.id),
                "polygon": [
                    [100, 100],
                    [150, 100],
                    [150, 150],
                    [100, 150],
                    [100, 100],
                ],
                "center": [125, 125],
                "url": self.z1.url,
                "image": {
                    "id": str(self.z1.image.id),
                    "width": self.z1.image.width,
                    "height": self.z1.image.height,
                    "path": self.z1.image.path,
                    "url": self.z1.image.url,
                    "s3_url": None,
                    "status": self.z1.image.status.value,
                }
            },
            "elements": [
                {
                    "id": str(self.page.id),
                    "name": "Volume 1, page 1r",
                    "type": "page",
                }
            ]
        })

    def test_region_create_requires_login(self):
        response = self.client.post(reverse('api:region-create', kwargs={'pk': str(self.page.id)}), {
            "analyser": "tasty",
            "class_name": "eggs",
            "confidence": 0.74,
            "polygon": [
                [13, 37],
                [13, 42],
                [42, 42],
                [42, 37],
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_region_create(self, get_tool_mock):
        get_tool_mock.return_value.slug = 'tasty'
        get_tool_mock.return_value.name = 'Tasty analyser'
        get_tool_mock.return_value.type = MLToolType.DLAnalyser
        get_tool_mock.return_value.version = 2

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:region-create', kwargs={'pk': str(self.page.id)}), {
            "analyser": "tasty",
            "class_name": "eggs",
            "confidence": 0.74,
            "polygon": [
                [13, 37],
                [13, 42],
                [42, 42],
                [42, 37],
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        region = Region.objects.get(id=response.json()['id'])
        self.assertEqual(region.ml_class.name, 'eggs')
        self.assertEqual(region.confidence, 0.74)
        self.assertEqual(region.source, self.ds)
        self.assertEqual(region.zone.polygon, Polygon([
            [13, 37],
            [13, 42],
            [42, 42],
            [42, 37],
            [13, 37],
        ]))
        self.assertEqual(region.zone.image, self.page.zone.image)
        self.assertEqual(region.elements.count(), 1)
        self.assertEqual(region.elements.get(), self.page)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_region_create_duplicate_zones(self, get_tool_mock):
        """
        Ensure creating a region with a zone that already exists more than once does not fail
        """
        get_tool_mock.return_value.slug = 'tasty'
        get_tool_mock.return_value.name = 'Tasty analyser'
        get_tool_mock.return_value.type = MLToolType.DLAnalyser
        get_tool_mock.return_value.version = 2

        polygon = Polygon.from_coords(1, 2, 3, 4)
        z1 = self.page.zone.image.zones.create(polygon=polygon)
        z2 = self.page.zone.image.zones.create(polygon=polygon)

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:region-create', kwargs={'pk': str(self.page.id)}), {
            "analyser": "tasty",
            "class_name": "eggs",
            "confidence": 0.74,
            "polygon": [
                [point.x, point.y]
                for point in polygon
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        region = Region.objects.get(id=response.json()['id'])
        self.assertEqual(region.elements.get(), self.page)
        # The region has been added to one of the two zones
        self.assertIn(region.zone, (z1, z2))

    def test_element_regions(self):
        response = self.client.get(reverse('api:element-regions', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()['results']
        self.assertListEqual(data, [
            {
                "id": str(self.r1.id),
                "source": {
                    "id": str(self.ds.id),
                    "type": "dla",
                    "slug": "tasty",
                    "name": "Tasty analyser",
                    "revision": "2",
                    "internal": True,
                },
                "ml_class": {
                    "id": str(self.ml_class.id),
                    "name": "bacon",
                },
                "confidence": 0.8,
                "zone": {
                    "id": str(self.z1.id),
                    "polygon": [
                        [100, 100],
                        [150, 100],
                        [150, 150],
                        [100, 150],
                        [100, 100],
                    ],
                    "center": [125, 125],
                    "url": self.z1.url,
                    "image": {
                        "id": str(self.z1.image.id),
                        "width": self.z1.image.width,
                        "height": self.z1.image.height,
                        "path": self.z1.image.path,
                        "url": self.z1.image.url,
                        "s3_url": None,
                        "status": self.z1.image.status.value,
                    }
                },
            },
            {
                "id": str(self.r2.id),
                "source": {
                    "id": str(self.ds.id),
                    "type": "dla",
                    "slug": "tasty",
                    "name": "Tasty analyser",
                    "revision": "2",
                    "internal": True,
                },
                "ml_class": {
                    "id": str(self.ml_class.id),
                    "name": "bacon",
                },
                "confidence": 0.5,
                "zone": {
                    "id": str(self.z2.id),
                    "polygon": [
                        [200, 200],
                        [250, 200],
                        [250, 250],
                        [200, 250],
                        [200, 200],
                    ],
                    "center": [225, 225],
                    "url": self.z2.url,
                    "image": {
                        "id": str(self.z2.image.id),
                        "width": self.z2.image.width,
                        "height": self.z2.image.height,
                        "path": self.z2.image.path,
                        "url": self.z2.image.url,
                        "s3_url": None,
                        "status": self.z2.image.status.value,
                    }
                },
            }
        ])

    def test_region_bulk_requires_login(self):
        response = self.client.post(reverse('api:region-bulk'), {
            "parent": str(self.page.id),
            "analyser": "tasty",
            "regions": [
                {
                    "class_name": "eggs",
                    "confidence": 0.74,
                    "polygon": [
                        [13, 37],
                        [13, 42],
                        [42, 42],
                        [42, 37],
                    ],
                },
                {
                    "class_name": "bacon",
                    "confidence": 0.85,
                    "polygon": [
                        [1, 1],
                        [1, 2],
                        [2, 2],
                        [2, 1],
                    ]
                }
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_region_bulk_create(self, get_tool_mock):
        get_tool_mock.return_value.slug = 'tasty'
        get_tool_mock.return_value.name = 'Tasty analyser'
        get_tool_mock.return_value.type = MLToolType.DLAnalyser
        get_tool_mock.return_value.version = 2

        self.assertEqual(self.page.regions.count(), 2)
        self.assertEqual(self.ml_class.regions.count(), 2)

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:region-bulk'), {
            "parent": str(self.page.id),
            "analyser": "tasty",
            "regions": [
                {
                    "class_name": "eggs",
                    "confidence": 0.74,
                    "polygon": [
                        [13, 37],
                        [13, 42],
                        [42, 42],
                        [42, 37],
                    ],
                },
                {
                    "class_name": "coffee",
                    "confidence": 0.85,
                    "polygon": [
                        [1, 1],
                        [1, 2],
                        [2, 2],
                        [2, 1],
                    ]
                }
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.page.regions.count(), 4)
        self.assertEqual(self.ml_class.regions.count(), 2)
        region1 = self.page.regions.get(ml_class__name='eggs')
        region2 = self.page.regions.get(ml_class__name='coffee')
        self.assertEqual(region1.confidence, 0.74)
        self.assertEqual(region2.confidence, 0.85)
        self.assertEqual(region1.source, self.ds)
        self.assertEqual(region2.source, self.ds)
        self.assertEqual(region1.zone.polygon, Polygon.from_coords(13, 37, 42-13, 42-37))
        self.assertEqual(region2.zone.polygon, Polygon.from_coords(1, 1, 1, 1))

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_region_bulk_update(self, get_tool_mock):
        get_tool_mock.return_value.slug = 'tasty'
        get_tool_mock.return_value.name = 'Tasty analyser'
        get_tool_mock.return_value.type = MLToolType.DLAnalyser
        get_tool_mock.return_value.version = 2

        self.assertEqual(self.page.regions.count(), 2)
        self.assertEqual(self.ml_class.regions.count(), 2)

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:region-bulk'), {
            "parent": str(self.page.id),
            "analyser": "tasty",
            "regions": [
                {
                    "class_name": "eggs",
                    "confidence": 0.74,
                    "polygon": [
                        [13, 37],
                        [13, 42],
                        [42, 42],
                        [42, 37],
                    ],
                },
                {
                    "class_name": "eggs",
                    "confidence": 0.85,
                    "polygon": [
                        [1, 1],
                        [1, 2],
                        [2, 2],
                        [2, 1],
                    ]
                }
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(self.page.regions.count(), 2)
        self.assertEqual(self.ml_class.regions.count(), 0)
        self.assertEqual(self.corpus.ml_classes.get(name='eggs').regions.count(), 2)
