from django.db.models import Q
from rest_framework.generics import ListAPIView, CreateAPIView, RetrieveUpdateAPIView
from rest_framework.serializers import ValidationError
from arkindex.project.aws import S3FileStatus
from arkindex.project.permissions import IsVerified
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.documents.models import Element, Corpus
from arkindex.images.models import Image
from arkindex.images.serializers import ImageSerializer, ImageUploadSerializer


class ImageCreate(CreateAPIView):
    """
    Create an image on the local server, ready for upload to S3
    """

    permission_classes = (IsVerified, )
    serializer_class = ImageUploadSerializer

    def perform_create(self, serializer):
        image_hash = serializer.validated_data.get('hash')
        server = serializer.validated_data.get('server')
        path = serializer.validated_data.get('path')
        existing_image = server.images.filter(
            Q(hash=image_hash) | Q(path=path)
        ).first()

        if existing_image:
            error = {
                'id': str(existing_image.id),
            }
            if existing_image.hash == image_hash:
                error['hash'] = ['Image with this hash already exists']
            if existing_image.path == path:
                error['path'] = ['Image with this path already exists']
            raise ValidationError(error)

        width = serializer.validated_data.get('width')
        height = serializer.validated_data.get('height')
        if width and height and self.request.user.is_internal:
            # Async tasks may send the image's width and height;
            # we will assume that the image was checked by the task
            serializer.validated_data['status'] = S3FileStatus.Checked

        return serializer.save()


class ImageRetrieve(RetrieveUpdateAPIView):
    """
    Get information about an image or trigger a check
    """

    permission_classes = (IsVerified, )
    serializer_class = ImageSerializer
    queryset = Image.objects.all()


class ImageElements(ListAPIView):
    """
    List elements linked to an image
    """

    serializer_class = ElementSlimSerializer
    openapi_overrides = {
        'operationId': 'ListImageElements',
        'description': 'List elements on an image',
        'security': [],
        'tags': ['images'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter by element type',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'hidden',
                'in': 'query',
                'description': 'Restrict to or exclude elements with hidden types; they are excluded by default',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
        ]
    }

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user),
            'zone__image_id': self.kwargs['pk'],
        }
        if 'type' in self.request.query_params:
            filters['type__slug'] = self.request.query_params['type']
        else:
            # Hide hidden elements if there isn't a type filter
            filters['type__hidden'] = False

        only_hidden = self.request.query_params.get('hidden')
        if only_hidden is not None:
            filters['type__hidden'] = only_hidden.lower() not in ('false', '0')

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        return Element.objects.filter(**filters) \
            .prefetch_related('corpus', 'zone__image__server', 'type') \
            .order_by('type__slug', 'name')
