from unittest.mock import patch, call
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureTestCase
import requests
import responses


class TestImagePerformCheck(FixtureTestCase):

    def setUp(self):
        self.img = self.imgsrv.images.get(path='img1')

    @responses.activate
    def test_status(self):
        """
        Test Image.perform_check fails on a HTTP error code
        """
        responses.add(responses.GET, 'http://server/img1/info.json', status=418)
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaises(requests.HTTPError):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_required_items(self):
        """
        Test Image.perform_check fails on a HTTP error code
        """
        base_payload = {
            '@id': 'http://server/img1',
            'width': 42,
            'height': 42,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        }
        for key in base_payload:
            payload = base_payload.copy()
            del payload[key]
            responses.add(responses.GET, 'http://server/img1/info.json', json=payload)
            with self.assertRaises(AssertionError, msg='Missing required properties'):
                self.img.perform_check(raise_exc=True, save=False)
            self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_check_id(self):
        """
        Test Image.perform_check ensures the image is on the expected server
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@id': 'http://wrongserver/img1',
            'width': 42,
            'height': 42,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.img.status = S3FileStatus.Unchecked
        with self.assertRaisesRegex(AssertionError, 'Image id does not start with server url'):
            self.img.perform_check(raise_exc=True, save=False)
        self.assertEqual(self.img.status, S3FileStatus.Error)

    @responses.activate
    def test_update_fields(self):
        """
        Test Image.perform_check updates the path, width and height
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@id': 'http://server/img1111',
            'width': 9000,
            'height': 8000,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })
        self.assertEqual(self.img.path, 'img1')
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.status = S3FileStatus.Unchecked

        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'img1111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

    @responses.activate
    def test_max_size(self):
        """
        Test Image.perform_check handles the IIIF maxWidth and maxHeight optional properties
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@id': 'http://server/img1',
            'width': 9000,
            'height': 8000,
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4500,
                    "maxHeight": 10000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 4500)
        self.assertEqual(self.img.height, 4000)

    @responses.activate
    def test_max_height_deduced(self):
        """
        Test Image.perform_check assumes maxHeight is equal to maxWidth
        when maxHeight is missing and maxWidth is defined
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@id': 'http://server/img1',
            'width': 1000,  # width < maxWidth
            'height': 8000,  # height > maxWidth
            'profile': [
                'http://iiif.io/api/image/2/level2.json',
                {
                    "maxWidth": 4000,
                }
            ],
        })
        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 500)
        self.assertEqual(self.img.height, 4000)

    @responses.activate
    def test_harvard_non_conform(self):
        """
        Test Image.perform_check falls back to properties from the Image Information object
        when they are not defined in the profile description object or when it is missing.
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@id': 'http://server/img1',
            'width': 9001,
            'maxWidth': 9000,
            'height': 9001,
            'maxHeight': 8999,
            'profile': ['http://iiif.io/api/image/2/level2.json'],
        })

        self.img.status = S3FileStatus.Unchecked
        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.width, 8999)
        self.assertEqual(self.img.height, 8999)

    @responses.activate
    def test_iiif_v1(self):
        """
        Test Image.perform_check supports a IIIF Image API 1.x payload
        """
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@context': 'http://library.stanford.edu/iiif/image-api/1.1/context.json',
            '@id': 'http://server/img1111',
            'width': 9000,
            'height': 8000,
            'profile': 'http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level0'
        })
        self.assertEqual(self.img.path, 'img1')
        self.assertEqual(self.img.width, 1000)
        self.assertEqual(self.img.height, 1000)
        self.img.status = S3FileStatus.Unchecked

        self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'img1111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

    @responses.activate
    def test_profile_uri_validation(self):
        responses.add(responses.GET, 'http://server/img1/info.json', json={
            '@context': 'http://library.stanford.edu/iiif/image-api/1.1/context.json',
            '@id': 'http://server/img1111',
            'width': 9000,
            'height': 8000,
            'profile': 'lol',
        })

        with patch('arkindex.images.models.logger') as logger_mock:
            self.img.perform_check(raise_exc=True, save=False)

        self.assertEqual(self.img.status, S3FileStatus.Checked)
        self.assertEqual(self.img.path, 'img1111')
        self.assertEqual(self.img.width, 9000)
        self.assertEqual(self.img.height, 8000)

        self.assertEqual(logger_mock.warning.call_count, 1)
        self.assertEqual(logger_mock.warning.call_args, call('Invalid IIIF profile URI'))
