from django.conf import settings
from rest_framework import serializers
from enum import Enum
from urllib.parse import quote, unquote
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.project.polygon import Point, Polygon


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        if not isinstance(obj, self.enum):
            obj = self.to_internal_value(obj)
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError('Value is not of type {}'.format(self.enum.__name__))


class MLToolField(serializers.CharField):

    def __init__(self, tool_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(tool_type, MLToolType)
        self.tool_type = tool_type

    def to_representation(self, obj):
        if isinstance(obj, str):
            try:
                obj = MLTool.get(settings.ML_CLASSIFIERS_DIR, self.tool_type, obj)
            except ValueError:
                return
        return obj.slug

    def to_internal_value(self, data):
        try:
            return MLTool.get(settings.ML_CLASSIFIERS_DIR, self.tool_type, data)
        except ValueError as e:
            raise serializers.ValidationError(str(e))


class PointField(serializers.ListField):
    child = serializers.IntegerField()
    min_length = 2
    max_length = 2

    def to_representation(self, point):
        return [point.x, point.y]

    def to_internal_value(self, coords):
        try:
            return Point(*super().to_internal_value(coords))
        except (AssertionError, TypeError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class PolygonField(serializers.ListField):
    child = PointField()
    min_length = 3

    def to_internal_value(self, data):
        try:
            return Polygon(super().to_internal_value(data))
        except (AssertionError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class IIIFPathField(serializers.CharField):
    """
    An IIIF image path field where special characters are escaped,
    including slashes.
    """

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        # If unquoting the value gives a different value, the value was already quoted
        if '/' in value or unquote(value) == value:
            value = quote(value, safe='')
        return value
